(**********************************************************************)
(*                                                                    *)
(*                        CPA beats ooCFA                             *)
(*                                                                    *)
(*       Frédéric Besson,      Inria Rennes - Bretagne Atlantique     *)
(*                                                                    *)
(*  Copyright Institut National de Recherche en Informatique et en    *)
(*  Automatique. All rights reserved                                  *)
(*                                                                    *)
(**********************************************************************)
(** CPA beats oo-CFA *)
Require Import Utf8.
Require Import List.
Require Import StdNotations.
Require Import Cool.
Require Import ICool.
Require Import PointsToAnalysis.
Require Import Fix.

(** We compare two analyses oo-CFA and CPA which do not use the same call-context [CC] 
   but share the same allocation context [AC]. *)
Parameter AC : Set.

(** and the same allocation function [alloc].
   Note that [alloc] is independent from the call context [CC]
   (A typical choice is to take as allocation context the allocation site)
   *)
Parameter alloc : PC -> Class -> (Var -> option AC) -> AC.

(** Here is the specification of ooCFA *)
Module ooCtx.
  (** oo-CFA keeps as call context [CC] the list of program points of the callers *)
  Definition CC := list PC.
  Definition bullet :CC := nil.
  Definition AC := AC.
  Definition alloc : CC -> PC -> Class -> (Var -> option AC) -> AC := fun cc => alloc.
  Definition push : CC -> PC -> Meth -> (Var -> option AC) -> CC := fun cc pc _ _ => pc::cc.
End ooCtx.

Module CPACtx.
  (** CPA keeps as call-context [CC] the abstract environment of the caller And the static method name of the call.
     *)
  Definition AC := AC.
  Definition CC := option (Meth * (Var -> option AC)).  
  Definition bullet : CC := None.

  Definition alloc : CC -> PC -> Class -> (Var -> option AC) -> AC := fun cc=> alloc.
  Definition push : CC -> PC -> Meth -> (Var -> option AC) -> CC := fun cc pc m e => Some (m,e). 
End CPACtx.

(** What is striking is that CPA is directly computable by fixpoint iteration.
   The abstract domain is finite (as soon as there are finitely many allocation contexts).
   This is not the case for oo-CFA – as soon as there is a recursion in the abstract program.
   Yet, CPA is (strictly) more precise than oo-CFA *)

(** For the purpose of the proof we shall introduce another analysis Ω-CFA.
   It is straightforward that Ω-CFA is more precise than both CPA and oo-CFA because it keeps more context than both.
*)
Module ΩCtx.
  Definition AC := AC.
  Definition CC := list (Meth * (Var -> option AC) * PC)%type.
  Definition bullet : CC := nil.

  Definition alloc : CC -> PC -> Class -> (Var -> option AC) -> AC := fun cc => alloc.
  Definition push : CC -> PC -> Meth -> (Var -> option AC) -> CC := fun cc pc m reg =>  (m,reg,pc)::cc.

End ΩCtx.

Module ooCFA := Analysis(ooCtx).
Module CPA := Analysis(CPACtx).
Module ΩCFA := Analysis(ΩCtx).
Import CPA.
Import ooCFA.
Import ΩCFA.


(** As a warm-up, we prove that ΩCFA is more precise than ooCFA.
   The structure of the proof is the following:
   - we provide a translation function τ_oo_Ω converting the result of ooCFA into the result of ΩCFA
   - we show that the translation can be seen as a concretisation function:
     [Lemma Ω_le_oo : forall P, ΩCFA.Ana P ⊆ τ_oo_Ω (ooCFA.Ana P).]
   - we show that the translation is precise
*)


Section ΩCFAooCFA.

  Definition α_Ω_oo (l : ΩCtx.CC) : ooCtx.CC :=
    List.map (@snd _ _) l.

  Definition τ_E_oo_Ω (E: ooCFA.CCEnv) : ΩCFA.CCEnv :=
    fun cc  => E (α_Ω_oo cc).
  
  Definition τ_oo_Ω (A: ooCFA.Atom -> Prop) : ΩCFA.Atom -> Prop :=
    ΩCFA.mkA (ooCFA.gHC A) (ooCFA.gHO A) (τ_E_oo_Ω (ooCFA.gEnv A)).

  (** The translation functions are monotonic *)
  Lemma τ_oo_Ω_mono : forall A A', A ⊆ A' -> τ_oo_Ω A ⊆ τ_oo_Ω A'.
  (* begin hide *)
  Proof.
    intros.
    unfold τ_oo_Ω.
    unfold ΩCFA.mkA.
    unfold leR ; intros.
    destruct x.
    eapply ooCFA.gHC_mono ; eauto.
    eapply ooCFA.gHO_mono ; eauto.
    unfold τ_E_oo_Ω in *.
    eapply ooCFA.leEnv_gEnv_mono ; eauto.
  Qed.
  (* end hide *)

  (* begin hide *)
  Lemma τ_oo_gHC : forall X, ΩCFA.gHC (τ_oo_Ω X) = ooCFA.gHC X.
  Proof. 
    reflexivity. 
  Qed.
    
  Lemma τ_oo_gHO : forall X, ΩCFA.gHO (τ_oo_Ω X) = ooCFA.gHO X.
  Proof. 
    reflexivity. 
  Qed.
  
  Lemma τ_oo_gEnv : forall X, ΩCFA.gEnv (τ_oo_Ω X) =  τ_E_oo_Ω (ooCFA.gEnv X).
  Proof.
    reflexivity.
  Qed.
  (* end hide *)

  Lemma ΩooEnvSpec : forall P cc pc r a E hc ho,
    ΩCFA.EnvSpec P hc ho (τ_E_oo_Ω E) cc pc r a
    → ooCFA.EnvSpec P hc ho E (α_Ω_oo cc) pc r a.
  Proof.
  (** Proof by case analysis over ΩCFA.EnvSpec *)
  (* begin hide *)
    intros.
    inv H.
    (* base case *)
    simpl. constructor.
           (* call *)
    change  (α_Ω_oo (push cc1 pc1 m e)) with
      (ooCtx.push  (α_Ω_oo cc1) pc1 m e).
    eapply ooCFA.ESI_call ; eauto.
    (* ret0 *)
    eapply ooCFA.ESI_ret0 ; eauto.
(*    constructor.*)
    (* ret' *)
    eapply ooCFA.ESI_ret' ; eauto.
    (* intra *)
    eapply ooCFA.intra_read_r0 ; eauto.
    eapply ooCFA.intra_read_r' ; eauto.
    eapply ooCFA.intra_copy_r0 ; eauto.
    eapply ooCFA.intra_copy_r' ; eauto.
    eapply ooCFA.intra_write ; eauto.
    eapply ooCFA.intra_alloc_r0 ; eauto.
    eapply ooCFA.intra_alloc_r' ; eauto.
    (* end hide *)
  Qed.

  Lemma ΩooObjectSpec : forall P hc E a f b,
    HObjectSpec P hc  (τ_E_oo_Ω E) a f b ->
    ooCFA.HObjectSpec P hc E a  f b.
  (* begin hide *)
  Proof.
    intros.
    inv H.
    eapply ooCFA.HOS_alloc ; eauto.
    eapply ooCFA.HOS_write ; eauto.
  Qed.
  (* end hide *)

  Lemma ΩooClassSpec : forall P  E a  b,
    HClassSpec P (τ_E_oo_Ω E) a  b ->
    ooCFA.HClassSpec P  E a b.
  (* begin hide *)
  Proof.
    intros.
    inv H.
    eapply ooCFA.HTS_alloc ; eauto.
  Qed.
  (* end hide *)

  Lemma Ω_le_oo : forall P, ΩCFA.Ana P ⊆ τ_oo_Ω (ooCFA.Ana P).
  Proof.
  (** The proof is by fixpoint transfer [lfp_corrct] *)
  (* begin hide *)
    unfold ΩCFA.Ana , ooCFA.Ana.
    unfold AnaI, ooCFA.AnaI. intro.
    change (ΩCFA.ASpecFI P bullet 0 e0) with (F _ (ΩCFA.ASpecFIDomKT P bullet 0 e0)).
    change (ooCFA.ASpecFI P ooCtx.bullet 0 ooCFA.e0) with (F _ (ooCFA.ASpecFIDomKT P ooCtx.bullet 0 ooCFA.e0)).
    eapply lfp_corrct ; eauto.
    apply τ_oo_Ω_mono.
    unfold Fc, Fa.
    simpl.
    intros. intro. intros.
    inv H ; constructor.
    eapply ΩooObjectSpec ; eauto.
    eapply ΩooClassSpec ; eauto.
    eapply ΩooEnvSpec ; eauto.
    (* end hide *)
  Qed.

  Definition α_Ωoo_iframe (ifr : ΩCFA.ISem.IFrame) : ooCFA.ISem.IFrame :=
    let (pc,e,cc) := ifr in
      [pc , e , α_Ω_oo cc].

  Definition α_Ωoo_istate (is : ΩCFA.ISem.IState) : ooCFA.ISem.IState :=
    let (h,s,ac) := is in
      [h , List.map α_Ωoo_iframe s , ac].
  
  Lemma Ω_EE_oo : forall P A, ΩCFA.γ_C P (τ_oo_Ω A) ⊆ ooCFA.γ_C P A.
  Proof.
  (** For each instrumented state Ωis ∈ ΩCFA.γ_C P (τ_oo_Ω A), we have to find oois ∈ ooCFA.γ_C P A.
     We prove that, oois = α_Ωoo_istate Ωis *)
  (* begin hide *)
    unfold ΩCFA.γ_C, ooCFA.γ_C.
    unfold leR. intros.
    destruct H.
    exists (α_Ωoo_istate x0).
    split.
    destruct H.
    inv H. constructor.
    clear Hh.
    induction Hf.
    constructor.
    rewrite τ_oo_gEnv in H.
    unfold τ_E_oo_Ω in H.
    unfold γ_E in H.
    unfold ooCFA.γ_E.
    apply H.
    simpl.
    eapply ooCFA.F_cons ; eauto.
    apply Hh.
    destruct H.
    subst.
    destruct x0 ; simpl.
    clear H.
    replace (eraseStack stack0) with (ooCFA.ISem.eraseStack (map α_Ωoo_iframe stack0)).
    reflexivity.
    induction stack0.
    reflexivity.
    simpl.
    rewrite IHstack0. 
    destruct a. simpl. reflexivity.
   (* end hide *)
  Qed.
  
  Lemma Ωbeatsoo : forall P,  ΩCFA.γ_C P (ΩCFA.Ana P) ⊆ ooCFA.γ_C P (ooCFA.Ana P).
  Proof.
  (** By [Ω_EE_oo], we have [ΩCFA.γ_C P (τ_oo_Ω (ooCFA.Ana P) ⊆ ooCFA.γ_C P (ooCFA.Ana P) *)
  (* begin hide *)
    intros.
    assert (HH := Ω_EE_oo P (ooCFA.Ana P)).
    (* end hide *)
    (** By [Ω_le_oo], we also have [ΩCFA.Ana P ⊆ τ_oo_Ω (ooCFA.Ana P)]. *)
    (* begin hide *)
    assert (HH' := Ω_le_oo P ).
    (* end hide *)
    (** The proof follows by monotonie of [γ_C] ans transitivity of [⊆] *)
    (* begin hide *)
    apply (γ_C_mono P) in HH'.
    eapply leR_trans ; eauto.
    (* end hide *)
  Qed.

End ΩCFAooCFA.

  
Section CPAΩCFA.
(** For proving that CPA is more precise than ΩCFA, we use a similar methodology. *)

(** In the previous case, the translation was based on an abstraction function.
   Here, it is not possible to map a CPACtx.CC to a single ΩCtx.CC.
   *)

Definition α_CPA_Ω (P: Program) (e: CPACtx.CC) : ΩCtx.CC -> Prop :=
  fun ctx => match e with
               | None => ctx = nil
               | Some (m,e) => exists  l , exists pco , nthp P pco = Some (Call m) /\ ctx = (m,e,pco)::l
             end.

Definition τ_E_Ω (P:Program) (E: ΩCtx.CC -> PC -> ΩCFA.AEnv) : CPACtx.CC -> PC -> CPA.AEnv := 
  fun e pc r a => exists ctx, α_CPA_Ω P e ctx /\ E ctx pc r a.

Lemma τ_E_Ω_mono : forall P E E', ΩCFA.leEnv E E' -> 
  forall cc pc r a, τ_E_Ω P  E cc pc r a -> τ_E_Ω P  E' cc pc r a.
(* begin hide *)
Proof.
  intros.
  unfold τ_E_Ω in *.
  destruct H0. exists x.
  intuition.
Qed.
(* end hide *)

(* begin hide *)
Definition τ_E_Ω' (P:Program) (E: ΩCtx.CC -> PC -> ΩCFA.AEnv) : CPACtx.CC -> PC -> CPA.AEnv := 
  fun e pc r a => 
    match e with
      | None => E nil pc r a
      | Some (m,e) => exists  l , exists pco ,
        nthp P pco = Some (Call m)  /\ (E ( (m,e,pco)::l) pc r a)
    end.

Lemma τ_E_Ω_Ω' : forall P E, CPA.leEnv (τ_E_Ω P E) (τ_E_Ω' P E).
  unfold CPA.leEnv.
  unfold τ_E_Ω, τ_E_Ω'.
  intros. destruct H.
  destruct cc0 ; intuition.
  destruct p.
  unfold α_CPA_Ω in H0.
  destruct H0 as [l [pco [HH HH1]]].
  subst. exists l ; exists pco. tauto.
  simpl in H0. congruence.
Qed.

Lemma τ_E_Ω'_Ω : forall P E, CPA.leEnv (τ_E_Ω' P E) (τ_E_Ω P E).
  unfold CPA.leEnv.
  unfold τ_E_Ω, τ_E_Ω'.
  intros. 
  destruct cc0 ; intuition.
  destruct p.
  unfold α_CPA_Ω.
  destruct H as [l [pco [HH HH1]]].
  exists ((m,o,pco)::l).
  split.
  exists l ; exists pco. tauto.
  auto.
  simpl. exists nil ; auto.
Qed.

Lemma τ_E_Ω_simpl : forall P E cc pc r a, τ_E_Ω P E cc pc r a <-> τ_E_Ω' P E cc pc r a.
Proof.
  intros ; split.
  apply τ_E_Ω_Ω'.
  apply τ_E_Ω'_Ω.
Qed.
(* end hide *)

Definition τ_Ω (P:Program) (Ab: ΩCFA.Atom -> Prop) : CPA.Atom -> Prop :=
  CPA.mkA (ΩCFA.gHC Ab) (ΩCFA.gHO Ab) (τ_E_Ω P (ΩCFA.gEnv Ab)).


Lemma τ_Ω_mono : forall P Ab Ab',
  Ab ⊆ Ab' ->  (τ_Ω P Ab) ⊆ (τ_Ω P Ab').
(* begin hide *)
Proof.
  unfold leR.
  unfold τ_Ω.
  intros.
  destruct x ; simpl in *.
  apply H. simpl ; auto.
  apply H. simpl ; auto.
  eapply τ_E_Ω_mono ; eauto.
  unfold leEnv. intros.
  eapply ΩCFA.leEnv_gEnv_mono ; eauto.
Qed.
(* end hide *)

(* begin hide *)
Lemma gHC_τ_Ω : forall P F, CPA.gHC (τ_Ω P F) = ΩCFA.gHC F.
Proof.
  reflexivity.
Qed.

Lemma gHO_τ_Ω : forall P F, CPA.gHO (τ_Ω P F) = ΩCFA.gHO F.
Proof.
  reflexivity.
Qed.

Lemma gEnv_τ_Ω : forall P F, CPA.gEnv (τ_Ω P F) =  τ_E_Ω P (ΩCFA.gEnv  F).
Proof.
  reflexivity.
Qed.

Lemma τ_Ω_Ecc : forall P A cc0 pc0 r a,
  τ_E_Ω P (EnvSpec P (gHC A) (gHO A) (gEnv A)) cc0 pc0 r a -> 
   τ_Ω P (ASpecF P A) (CPA.Ecc cc0 pc0 r a) .
Proof.
  intros.
  unfold τ_Ω.
  unfold CPA.mkA.
  eapply τ_E_Ω_mono ; eauto.
  unfold leEnv.
  intros.
  unfold gEnv. constructor.
  auto.
Qed.
(* end hide *)


(* begin hide *)
Ltac by_length  H :=
  match type of H with 
    @eq ?A ?E ?E'  => let Hlen := fresh "hlen" in 
      assert (Hlen : length E = length E') by congruence ;
        simpl in Hlen ; try rewrite app_length in Hlen ; simpl in Hlen ;
          try subst ; apply False_ind ; try omega
  end.
(* end hide *)


(** In the following, we prove properties relative to the initial conditions of ΩCFA 
   namely [EnvSpecIFxLength] [EnvSpecIFxRoot], [EnvSpecIFxIn] and [EnvSpecIFxPC]. *)


(** By construction, all the contexts computed by the analysis are suffixes of the initial context.
   As a consequence, these are necessarily longer *)

Lemma EnvSpecIFxLength : forall P l e HC HO pco l' pc r a,
  EnvSpecIFx P l  pco e HC HO  l' pc r a -> length l <= length l'.
Proof.
(** The proof is by induction over the iterates of the fixpoint computation *)  
(* begin hide *)
  intros.
  unfold EnvSpecIFx in *.
  unfold mkCCEnv in *.
  destruct H. 
  revert H. revert l' pc0 r a.
  induction x ; intros.
  (* end hide *)
  (**  The base case is absurd *)
  (* begin hide *)
  intros. simpl in H. tauto.
  (* end hide *)
  (** For the inductive case, we perform a case analysis *)
  (* begin hide *)
  simpl in *.  inv H. 
  (** For [ESI_init], the context is the initial context. *)
  (* begin hide *)
  omega.
  (* end hide *)
  (** For a call, the context is a [push] from a previously computed (longer) context and 
     the  proof follows by induction hypothesis. *)
  (* begin hide *)
  unfold mem, mkCCEnv in H1.
  unfold ΩCtx.push.  simpl.
  assert (length l <= length cc0). eapply IHx ; apply (H1 v0).
  omega.
  (* end hide *)
  (** For a return, we apply directly the induction hypothesis – we return to a previously computed context *)
  (* begin hide *)
  eapply IHx ; apply (H3 v0).
  eapply IHx ; apply (H1 v0).
  (* end hide *)
  (** None of the interprocedural statement modifies the context ; the proof follows by induction hypothesis *)
  (* begin hide *)
  eapply IHx ; apply (H1 v0).  
  eapply IHx ; apply (H1 v0).
  eapply IHx ; apply (H1 v0).
  eapply IHx ; apply (H2 v0).
  eapply IHx ; apply (H1 v0).
  eapply IHx ; apply (H1 v0).
  eapply IHx ; apply (H1 v0).
  (* end hide *)
Qed.


(** Roughly speaking, [EnvSpecIFxRoot] states that whatever the initial call-context,
   the result of the analysis is the same for suffixes of this call context *)

Lemma EnvSpecIFxRoot : forall P l e HC HO pco  ls l' pc  r a,
  EnvSpecIFx P l  pco e HC HO  (ls ++ l) pc r a -> 
  EnvSpecIFx P l' pco e HC HO  (ls ++ l') pc r a.
Proof.
  (** The proof is by induction over the iterates of the fixpoint computation *)
  (* begin hide *)
  intros.
  unfold EnvSpecIFx in *.
  unfold mkCCEnv in *.
  destruct H. exists x.
  revert H. revert l' ls pc0 r a.
  induction x ; intros.
  (* end hide *)
  (**  The base case is absurd *)
  (* begin hide *)
  intros. simpl in H. tauto.
  (* end hide *)
  (** For the inductive case, we perform a case analysis *)
  (* begin hide *)
  simpl in *.  inv H. 
  (** For [ESI_init], we have that ls is necessarily nil
     and the proof follows by [ESI_init]. *)
  (* begin hide *)
  destruct ls.
  simpl. constructor. 
  by_length H0.
  (* end hide *)
  (** For a call, the call context [cc0] is necessarily a suffix of [ls++l].
     As a result, the proof follows by induction hypothesis *)
  (* begin hide *)
  intros.
  unfold ΩCtx.push in H0. 
  destruct ls. 
  simpl in *.
  (* This is not a prefix *)
  subst. clear IHx.
  apply False_ind.
  assert (Hr0 := H2 v0).
  unfold mkCCEnv in Hr0.
  assert  (EnvSpecIFx P ((m,e1,pc1)::cc0) pco e HC0 HO0 cc0 pc1 v0 (e1 v0)).
  exists x. auto.
  apply EnvSpecIFxLength in H. simpl in H. omega.
  (* call IC *)
  simpl in H0. inv H0.
  simpl.
  eapply ΩCFA.ESI_call ; eauto.
  unfold mem. intro.
  unfold mkCCEnv.
  eapply IHx. apply H2.
  (* end hide *)
  (** For a return and intra-procedural statements the proof follows by induction hypothesis. *)
  (* begin hide *)
  (* call / ret *)
  intros.
  eapply ESI_ret0 with (e:=e1);  eauto. 
   unfold mem. intro. unfold mkCCEnv.
   unfold push.
  change ((m, e1, pc') :: ls ++ l') with
    (((m, e1, pc') :: ls) ++ l').
  eapply IHx ; eauto.
  apply H2.
  intro.
  eapply IHx ; eauto. apply H3.
  (**)
  intros.
  eapply ESI_ret' ; eauto. intro. apply IHx ; apply H1.
  (* intra *)
  intros. eapply intra_read_r0 ; eauto.
  intro. apply IHx ; apply H1.
  intros. eapply intra_read_r' ; eauto.
  intro. apply IHx ; apply H1.
  intros. eapply intra_copy_r0 ; eauto.
  intro. apply IHx ; apply H1.
  intros. eapply intra_copy_r' ; eauto.
  intro. apply IHx ; apply H2.
  intros. eapply intra_write ; eauto.
  intro. apply IHx ; apply H1.
  unfold alloc.
  eapply intra_alloc_r0 ; eauto.
  intro. apply IHx ; apply H1 ; auto.
  eapply intra_alloc_r' ; eauto.
  intro. apply IHx ; apply H1 ; auto.
  (* end hide *)
Qed.

Lemma EnvSpecFxIn : forall P HT HO m o pco l ll p r a,
  EnvSpecFx P  HT HO (ll++(m, o, pco) :: l) p r a ->
  o ∈ (EnvSpecFx P  HT HO l pco).
Proof.
(** The proof is by induction over the iterates of the fixpoint computation *)
  (* begin hide *)
  intros.
  unfold EnvSpecIFx in *.
  unfold mkCCEnv in *.
  unfold mem.
  destruct H.
  revert H.
  revert ll m o pco p l r a.
  induction x ; simpl.
  (* bc *)
  intros. tauto.
  (* ic *)
  intros. inv H.
  unfold bullet in H1. by_length H1.
  (* call *)
  unfold push in H0.
  destruct ll. simpl in H0. inv H0.
  exists x. apply H2.
  simpl in H0. inv H0.
  apply IHx with (ll:= ll) (m:=m) (o:=o) (pco:=pco) (p:= pc0) (r:=v) (a:= e v); eauto.
  apply H2.
  (* call - ret *)
  revert v.
  unfold mem, mkCCEnv in H3.
  apply IHx with (r:=v0) (1:= H3 v0). 
  (* call - ret *)
  intros.
  apply IHx with (r:=v0) (1:= H1 v0).
  (* read *)
  intros. apply IHx with (r:= v0) (1:= H1 v0).
  (* read *)
  intros. apply IHx with (r:= v0) (1:= H1 v0).
  intros. apply IHx with (r:= v0) (1:= H1 v0).
  intros. apply IHx with (r:= v0) (1:= H2 v0).
  intros. apply IHx with (r:= v0) (1:= H1 v0).
  intros. apply IHx with (r:= v0) (1:= H1 v0).
  intros. apply IHx with (r:= v0) (1:= H1 v0).
  (* end hide *)
Qed.

(* begin hide *)
Lemma EnvSpecIFx_pf : forall P cc pc e HC HO,
  leEnv
  (EnvSpecI P cc pc e HC HO  (EnvSpecIFx P cc pc e HC HO))
  (EnvSpecIFx P cc pc e HC HO).
Proof.
  unfold leEnv.
  intros.
  eapply lfpK_pf ; eauto.
  unfold mono ; intros.
  eapply EnvSpecF_mono ; eauto.
  unfold leHC;auto.
  unfold leHO;auto.
  apply EnvSpecF_cont ; auto.
Qed.
(* end hide *)
  
Lemma EnvSpecIFxPC : forall P HT HO m o pco pco' l ll pc r a,
  nthp P pco = nthp P pco' ->
  EnvSpecFx P HT HO (ll++(m, o, pco) :: l) pc r a ->
  EnvSpecIFx P l pco' o HT HO (ll++ (m,o,pco') :: l) pc r a.
Proof.
  (** The proof is by induction over the iterates of [EnvSpecFx] *)
  (* begin hide *)
  intros until a.  intro Nth.
  intros.
  revert Nth.
  destruct H.
  revert H.
  revert ll m o pco.
  revert pc0 r a.
  induction x ; simpl.
  tauto.
  (* end hide *)
  (** and case analysis over [EnvSpecI]  *)
  (* begin hide *)
  intros.
  inv H.
  (* end hide *)
  (** The base case is trivially absurd *)
  (* begin hide *)
  unfold ΩCtx.bullet in H1. by_length H1.
  (* end hide *)
  (** To solve the other subgoals, we unroll the definition of [EnvSpecIFx] and use the induction hypothesis *)
  (* begin hide *)
  (* call *)
  intros. unfold push in H0.
  destruct ll. simpl in H0.
  inv H0.
  simpl.
  apply EnvSpecIFx_pf.
  apply ESI_call;auto.  congruence.
  unfold mem. intro. apply EnvSpecIFx_pf. eapply ESI_init.
  inv H0.
  apply EnvSpecIFx_pf.
  simpl.
  eapply ESI_call ; eauto.  intro.  eapply IHx ; eauto. apply (H2 v).
  (* call - ret *)
  intros. 
  apply EnvSpecIFx_pf.
  apply ESI_ret0 with (pc:=pc1) (e:=e) (m:=m0); eauto.
  intro. unfold push. 
  change ((m0, e, pc') :: ll ++ (m, o, pco') :: l) with
    ( ((m0, e, pc') :: ll) ++ (m, o, pco') :: l).
  apply IHx with (pco:=pco);auto.
  apply (H2 v).
  intro.
  apply IHx with (pco:=pco);auto.
  apply (H3 v).
  (* call - ret *)
  intros.
  apply EnvSpecIFx_pf.
  eapply ESI_ret' ; eauto.
  intro.
  apply IHx with (pco:=pco);auto.
  apply H1.
  (* read *)
  intros.   apply EnvSpecIFx_pf. eapply intra_read_r0 ; eauto. 
  intro.  apply IHx with (pco:=pco); auto. apply H1. 
  intros.   apply EnvSpecIFx_pf. eapply intra_read_r' ; eauto. 
  intro.  apply IHx with (pco:=pco) ; auto. apply H1. 
  intros.   apply EnvSpecIFx_pf. eapply intra_copy_r0 ; eauto. 
  intro.  apply IHx with (pco:=pco) ; auto. apply H1. 
  intros.   apply EnvSpecIFx_pf. eapply intra_copy_r' ; eauto. 
  intro.  apply IHx with (pco:=pco) ; auto. apply H2. 
  intros.   apply EnvSpecIFx_pf. eapply intra_write  ; eauto. 
  intro.  apply IHx with (pco:=pco) ; auto. apply H1. 
  intros.   apply EnvSpecIFx_pf. unfold alloc. eapply intra_alloc_r0  ; eauto. 
  intro.  apply IHx with (pco:=pco) ; auto. apply H1. 
  intros.   apply EnvSpecIFx_pf. eapply intra_alloc_r'  ; eauto. 
  intro.  apply IHx with (pco:=pco) ; auto. apply H1. 
 (* end hide *)
Qed.


Lemma EnvSpecSwap : forall P HC HO m o pco pco' l l'  pc' e v v',
  nthp P pco = nthp P pco'-> 
  EnvSpecFx P HC HO ((m, o, pco') :: l') pc' v' (e v') -> 
  EnvSpecFx P HC HO ((m, o, pco) :: l) pc' v (e v) -> 
   EnvSpecFx P HC HO ((m, o, pco) :: l) pc' v' (e v').
Proof.
  (** By [EnvSpecIFxPC] we obtain [EnvSpecIFx P l' pco o HT HO ((m, o, pco) :: l') pc' v (e v)] *)
  (* begin hide *)
  intros.
  change (((m, o, pco') :: l')) with (nil ++ (m, o, pco') :: l') in H0.
  dup H0 HH2.
  apply EnvSpecIFxPC with (pco' := pco) in H0;auto ; try congruence.
  simpl in H0.
  (* end hide *)
  (** By [EnvSpecIFxRoot], we obtain [EnvSpecIFx P l pco o HT HO0 ((m, o, pco) ::  l) pc' v (e v)] *)
  (* begin hide *)
  change (((m, o, pco) :: l')) with (((m,o,pco)::nil) ++ l') in H0.
  apply EnvSpecIFxRoot with (l':= l) in H0.
  (* end hide *)
  (** By [EnvSpecFxIn] we obtain [o ∈ EnvSpecFx P HT HO l pco] *)
  (* begin hide *)
  change (((m, o, pco) :: l)) with (nil++((m,o,pco):: l)) in H1.
  apply EnvSpecFxIn in H1.
  (* end hide *)
  (** We conclude by [EnvSpecFxTrans] *)
  (* begin hide *)
  eapply EnvSpecFxTrans with (1:=H1); eauto.
  (* end hide *)
Qed.


Lemma α_EnvSpecFx : forall P HC HO cc ctx ctx' pc' v v' e,
  α_CPA_Ω P cc ctx  -> 
  EnvSpecFx P HC HO ctx pc' v (e v) -> 
  α_CPA_Ω P cc ctx' -> 
  EnvSpecFx P HC HO ctx' pc' v' (e v') -> 
   EnvSpecFx P HC HO ctx pc' v' (e v').
Proof.
  (** The proof is by case analysis over [cc] using [EnvSpecSwap] *)
  (* begin hide *)
  intros.
  destruct cc0 ; simpl in *.
  destruct p.
  destruct H as [l [pco [Hc Hctx]]].
  destruct H1 as [l' [pco' [Hc' Hctx']]].
  subst.
  eapply EnvSpecSwap with (2:= H2) (3:= H0); eauto. 
  congruence.
  subst.
  auto.
  (* end hide *)
Qed.


Lemma mem_τ : forall P HT HO cc pc' e,
  e ∈ τ_E_Ω P (EnvSpecFx P HT HO) cc pc' -> 
  exists l, (α_CPA_Ω P cc l)  /\ ( e ∈ (EnvSpecFx P HT HO l pc')) .
Proof.
  (** The proof follows from [α_EnvSpecFx] *)
  (* begin hide *)
  intros.
  unfold τ_E_Ω in H.
  unfold mem in H.
  destruct (H v0).
  exists x.
  intuition.
  intro. 
  destruct (H v).
  destruct H0.
  eapply α_EnvSpecFx ; eauto.
  (* end hide *)
Qed.
  

Lemma CPAΩ_EnvSpec : forall P HT HO, CPA.leEnv (CPA.EnvSpecFx P  HT HO)  (τ_E_Ω P (ΩCFA.EnvSpecFx P HT HO) ).
Proof.
  (** The proof is by induction over [CPA.EnvSpecFx] and case analysis over [CPA.EnvSpecF]. *)
  (* begin hide *)
  intros.
  unfold CPA.leEnv. intros.
  destruct H. revert H.
  revert cc0 pc0 r a.
  induction x;simpl ; intros.
  (* end hide *)
  (** The base case is trivially absurd *)
  (* begin hide *)
  tauto.
  inv H.
  (* end hide *)
  (** The case [ESI_init] is solved by definition of [EnvSpecI] *)
  (* begin hide *)
  rewrite τ_E_Ω_simpl. simpl.
  apply EnvSpecIFx_pf. constructor.
  (* end hide *)
  (** The call case is solved by induction hypothesis using [mem_τ']. *)
  (* begin hide *)
  unfold CPACtx.push.
  assert (e ∈ (τ_E_Ω P (EnvSpecFx P HT HO0)) cc1 pc1).
   intro. apply IHx. apply H1.
  rewrite τ_E_Ω_simpl.
  unfold τ_E_Ω'. 
  apply mem_τ in H. destruct H. exists x0. exists pc1.
  split ; auto.
  apply EnvSpecIFx_pf.
  eapply ESI_call ; intuition.
  (* end hide *)
  (** By induction hypothesis and [mem_τ] we exhibit a caller environment. *)
  (* begin hide *)
  assert (H :e ∈ (τ_E_Ω P (EnvSpecFx P HT HO0)) cc0 pc').
   intro. apply IHx. apply H3.
  dup H H'.
  apply mem_τ in H.
  destruct H as [l [Ha He]].
  unfold τ_E_Ω.
  exists l.
  split ; auto.
  apply EnvSpecIFx_pf.
  apply ESI_ret0 with (m :=m) (e:=e) (pc:=pc1) ; auto.
  unfold push.
  intro.
  assert (Hr : er ∈ (τ_E_Ω P (EnvSpecFx P HT HO0)) (CPACtx.push cc0 pc' m e) pc1).
   intro. apply IHx. apply H2.   
  apply mem_τ in Hr.
  destruct Hr. destruct H.
  destruct H as [ll [pcol [Hn' He']]] ; subst.
  eapply EnvSpecFxTrans with (1:= He); eauto.
  change AC with CPACtx.AC.
  change ((m,e,pc') ::l) 
    with (((m,e,pc')::nil)++(l)).
  apply EnvSpecIFxRoot with (l:= ll).
  simpl.
  change AC with CPACtx.AC.
  change ((m,e,pc')::ll) with (nil ++ (m,e,pc')::ll). 
  apply EnvSpecIFxPC with (pco:=pcol); eauto.
  congruence.
  (* end hide *)
  (** By [ESI_ret0], we have to prove that the context of the return
     is [((m, e, pc') :: l)] *)
  (** This is done by [EnvSpecFxTrans] , [EnvSpecIFxRoot] and [EnvSpecIFxPC] *)
  (** Other cases keep the context unchanged. Hence, they are discharged by induction hypothesis *)
  (* begin hide *)
  (* call-ret '*)
  assert (Her : er ∈ (τ_E_Ω P (EnvSpecFx P HT HO0)) cc0 pc').
   intro. apply IHx. apply H1.
  apply mem_τ in Her.
  destruct Her as [l [Ha HH]].
  exists l ; intuition.
  apply EnvSpecIFx_pf.
  eapply ESI_ret' ; eauto.
  (* read *)
  assert (Her : e ∈ (τ_E_Ω P (EnvSpecFx P HT HO0)) cc0 pc1).
   intro. apply IHx. apply H1.
  apply mem_τ in Her.
  destruct Her as [l [Ha HH]].
  exists l ; intuition.
  apply EnvSpecIFx_pf.
  eapply intra_read_r0 ; eauto.
  (**)
  assert (Her : e ∈ (τ_E_Ω P (EnvSpecFx P HT HO0)) cc0 pc1).
   intro. apply IHx. apply H1.
  apply mem_τ in Her.
  destruct Her as [l [Ha HH]].
  exists l ; intuition.
  apply EnvSpecIFx_pf.
  eapply intra_read_r' ; eauto.
  (* copy *)
  assert (Her : e ∈ (τ_E_Ω P (EnvSpecFx P HT HO0)) cc0 pc1).
   intro. apply IHx. apply H1.
  apply mem_τ in Her.
  destruct Her as [l [Ha HH]].
  exists l ; intuition.
  apply EnvSpecIFx_pf.
  eapply intra_copy_r0 ; eauto.
  (**)
  assert (Her : e ∈ (τ_E_Ω P (EnvSpecFx P HT HO0)) cc0 pc1).
   intro. apply IHx. apply H2.
  apply mem_τ in Her.
  destruct Her as [l [Ha HH]].
  exists l ; intuition.
  apply EnvSpecIFx_pf.
  eapply intra_copy_r' ; eauto.
  (* write *)
  assert (Her : e ∈ (τ_E_Ω P (EnvSpecFx P HT HO0)) cc0 pc1).
   intro. apply IHx. apply H1.
  apply mem_τ in Her.
  destruct Her as [l [Ha HH]].
  exists l ; intuition.
  apply EnvSpecIFx_pf.
  eapply intra_write ; eauto.
  (* alloc *)
  assert (Her : es ∈ (τ_E_Ω P (EnvSpecFx P HT HO0)) cc0 pc1).
   intro. apply IHx. apply H1.
  apply mem_τ in Her.
  destruct Her as [l [Ha HH]].
  exists l ; intuition.
  apply EnvSpecIFx_pf.
  eapply intra_alloc_r0 ; eauto.
  (**)
  assert (Her : es ∈ (τ_E_Ω P (EnvSpecFx P HT HO0)) cc0 pc1).
   intro. apply IHx. apply H1.
  apply mem_τ in Her.
  destruct Her as [l [Ha HH]].
  exists l ; intuition.
  apply EnvSpecIFx_pf.
  eapply intra_alloc_r' ; eauto.
 (* end hide *)
Qed.

Lemma CPAΩ_mem : forall P HT HO,
  forall  cc pc es, es ∈  (CPA.EnvSpecFx P HT HO cc pc) ->  
    ∃ l, es ∈  (ΩCFA.EnvSpecFx P HT HO) l pc.
(* begin hide *)
Proof.
  intros.
  assert (es  ∈ (τ_E_Ω P (ΩCFA.EnvSpecFx P HT HO0) cc0 pc0)).
  intro.
  apply CPAΩ_EnvSpec. apply H. clear H.
  apply mem_τ in H0.
  destruct H0 as [l  [HH1 HH2]].
  exists l. auto.
Qed.
(* end hide *)

Lemma CPAΩ_HClass : forall P HT HO,
  forall a t,  CPA.HClassSpec P (CPA.EnvSpecFx P HT HO) a t ->  ΩCFA.HClassSpec P (ΩCFA.EnvSpecFx P HT HO) a t.
Proof.
  (** By definition of [ΩCFA.HClassSpec] using  [CPAΩ_mem]. *)  
  (* begin hide *)
  intros.
  inv H.
  apply CPAΩ_mem in H1.
  destruct H1 as [l H1].
  eapply ΩCFA.HTS_alloc;eauto. 
  (* end hide *)
Qed.

Lemma CPAΩ_HObject : forall P HT HO,
  forall a f b,  CPA.HObjectSpec P HT (CPA.EnvSpecFx P HT HO) a f b ->  ΩCFA.HObjectSpec P HT (ΩCFA.EnvSpecFx P HT HO) a f b.
Proof.
  (** By definition of [ΩCFA.HObjectSpec] using [CPAΩ_mem]. *)
  (* begin hide *)
  intros.
  inv H.
  apply CPAΩ_mem in H1.
  destruct H1 as [l H1].
  eapply ΩCFA.HOS_alloc;eauto. 
  apply CPAΩ_mem in H1.
  destruct H1 as [l H1].
  eapply ΩCFA.HOS_write ; eauto.
(* end hide *)
Qed.


Lemma CPAΩCFA_AnaI' : forall P, CPA.AnaI' P CPACtx.bullet 0 CPA.e0 ⊆ τ_Ω P (AnaI' P bullet 0 e0).
Proof.
  (* The function [τ_Ω] plays the role of a concretisation function. 
     As a result,  we apply the lemma [lfp_corrct]  *)
  (* begin hide *)
  intros.
  unfold CPA.AnaI', AnaI'.
  change (ΩCFA.ASpecFI' P bullet 0 e0) with (F _ (ΩCFA.ASpecFI'DomKT P bullet 0 e0)).
  change (CPA.ASpecFI' P CPACtx.bullet 0 CPA.e0) with (F _ (CPA.ASpecFI'DomKT P CPACtx.bullet 0 e0)).
  apply lfp_corrct.
  apply τ_Ω_mono.
  simpl. 
  intros. intro. intros.
  (* end hide *)
  (* The proof is by case analysis over the definition of [CPA.ASpecFI'] *)
  (* begin hide *)
   inv H.
  (* end hide *)
  (* For [CPA.HObjectSpec], the proof follows from [CPAΩ_HObject] *)
  (* begin hide *)
  simpl in *.
  constructor.
  eapply CPAΩ_HObject ; eauto.
  (* end hide *)
  (* For [CPA.HClassSpec], the proof is by [CPAΩ_HClass] *)
  (* begin hide *)
  simpl. constructor.
  eapply CPAΩ_HClass ;eauto.
  (* end hide *)
  (* The [CPA.EnvSpecI] case requires some more care *)
  (* begin hide *)
  simpl.
  assert (Hclean : CPA.EnvSpec P (gHC e) (gHO e)
         (CPA.EnvSpecFx P (gHC e) (gHO  e)) cc0 pc0    r a).
   eapply CPA.EnvSpec_mono; eauto.
   unfold CPA.leHC ;auto.
   unfold CPA.leHO; auto.
   unfold CPA.leEnv ; auto. clear H0.
  (* end hide *)
  (* We exploit the fact that [CPA.EnvSpecFx] is a fixpoint.
     As a result, we can use [CPAΩ_EnvSpec] and apply the monotonie of [τ_E_Ω] *)
  (* begin hide *)
  apply CPA.EnvSpecFx_pf in Hclean.
  eapply CPAΩ_EnvSpec in Hclean.
  eapply τ_E_Ω_mono ; eauto.
  unfold leEnv. unfold gEnv. intros.
  constructor.
  apply  EnvSpecFx_fix in H.
  eapply EnvSpec_mono ; eauto.
  unfold leHC, gHC, mkA ; auto.
  unfold leHO, gHO, mkA ; auto.
  unfold leEnv, gHO, mkA ; auto.
  (* end hide *)
Qed.

Lemma CPAΩCFA : forall P, CPA.Ana P ⊆  τ_Ω P  (ΩCFA.Ana P).
(* begin hide *)
Proof.
  intros. 
  eapply leR_trans with (B:= CPA.AnaI' P CPACtx.bullet 0 CPA.e0) ; eauto.
  apply CPA.AnaAnaI'.
  intro. intros.
  apply CPAΩCFA_AnaI' in H.
  eapply τ_Ω_mono ; eauto.
  apply AnaI'Ana.
Qed.
(* end hide *)

Lemma γ_E_CPAΩ : forall  ae ac e, 
  CPA.γ_E  ac ae e -> 
  γ_E ac ae  e.
Proof.
  auto.
Qed.

Lemma γ_E_τ_bullet : forall P HT HO ac pc e,
 γ_E ac (τ_E_Ω P (EnvSpecFx P HT HO) CPACtx.bullet pc) e -> 
   γ_E ac (EnvSpecFx P HT HO bullet pc) e.
(* begin hide*)
Proof.
  unfold γ_E.
  intros.
  unfold mem in *.
  intro.
  generalize (H v).
  rewrite τ_E_Ω_simpl. simpl. auto.
Qed.
(* end hide *)


Lemma γ_F_τ_Ω : forall P HT HO ac l, 
  CPA.γ_F P (τ_E_Ω P (ΩCFA.EnvSpecFx P HT HO)) ac l -> 
   ∃l' : list IFrame,
   eraseStack l' = CPA.ISem.eraseStack l ∧ γ_F P (ΩCFA.EnvSpecFx P HT HO) ac l'.
Proof.
  (* The proof is by induction over the definition of [CPA.γ_F] *)
  (* begin hide *)
  intros.
  induction H.
  (* bc *)
  apply γ_E_CPAΩ in H.
  exists ((MkF pc0 e nil)::nil).
  split. reflexivity.
  eapply F_init.
  apply γ_E_τ_bullet ; auto.
  (* ic *)
  destruct IHγ_F as [l' [Herase Hg]].
  unfold CPACtx.push in H2.
  destruct l' ; simpl in Herase.
  discriminate. inv Herase.
  destruct i.
  exists ((MkF pc' e' (push cc1 pc0 m (abs e ac))) :: (MkF pc1 env0 cc1) :: l').
  split.
  simpl. simpl in H4. congruence.
  simpl in H4. inv H4.
  eapply F_cons ; eauto.
  unfold push.
  unfold γ_E, CPA.γ_E in *.
  apply mem_τ in H2.
  destruct H2 as [ll [Ha Hin]] .
  destruct Ha as [ll' [pco [Hnt Heq]]].
  subst.
  intro.
  apply γ_F_γ_E in Hg.
  unfold γ_E in Hg.
  eapply EnvSpecFxTrans with (1:= Hg); eauto.
  unfold PC.
  change ((m, abs e ac, pc0):: cc1) with (((m,abs e ac,pc0)::nil) ++cc1).
  eapply EnvSpecIFxRoot with (l:=ll'); eauto.
  simpl.
  unfold PC.
  change ((m, abs e ac, pc0):: ll') with (nil++(m,abs e ac,pc0)::ll').
  eapply EnvSpecIFxPC with (pco:=pco); eauto. 
  simpl. 
  congruence.
  (* end hide *)
Qed.
  
Lemma γ_τ_Ω : forall P, CPA.γ_C P (τ_Ω P  (ΩCFA.AnaI' P bullet 0 e0)) ⊆ γ_C P (ΩCFA.Ana P).
Proof.
  (* begin hide *)
  intros.
  unfold CPA.γ_C, γ_C.
  intro. intros.
  destruct H.
  destruct H. subst.
  inv H.
  assert (HH :CPA.γ_F P (τ_E_Ω P (ΩCFA.EnvSpecFx P (gHC (AnaI' P bullet 0 e0)) (gHO (AnaI' P bullet 0 e0)))) ac l).
   eapply CPA.γ_F_mono ; eauto.
   unfold τ_Ω , CPA.leEnv.
   intros until 0.
   unfold CPA.gEnv, CPA.mkA.
   eapply τ_E_Ω_mono . 
   apply gEnvAna'EnvSpecFx.
  rename HH into H.
  apply γ_F_τ_Ω in H.
  destruct H as [l' [He Hff]].
  exists (MkS  h l' ac).
  split.
  constructor.
  eapply γ_F_mono ; eauto.
  unfold leEnv ; intros.
  apply EnvSpecFxgEnvAna' in H.
  eapply leEnv_gEnv_mono ; eauto.
  apply AnaI'Ana.
  clear Hf He Hff.
  unfold CPA.γ_H in Hh.
  unfold γ_H ; intros.
  generalize (Hh a) ; clear Hh ; intro Hh.
  destruct (h a). destruct o. destruct (ac a).
  destruct Hh.
  split.
  unfold CPA.gHC, τ_Ω in H. simpl in H.
  eapply gHC_mono ; eauto.
  apply AnaI'Ana.
  intros.
  generalize (H0 f).
  unfold CPA.gHO , τ_Ω.
  simpl.
  eapply gHO_mono ; eauto.
  apply AnaI'Ana. auto.
  auto.
  (**)
  unfold CPA.ISem.EE , EE.
  rewrite He. reflexivity.
(* end hide *)
Qed.
   
Lemma CPAbeatsΩCFA : forall P, CPA.γ_C P (CPA.Ana P) ⊆ ΩCFA.γ_C P (ΩCFA.Ana P).
Proof.
  (* We have [CPAΩCFA]. By monotonie of [CPA.γ_C] we obtain 
     [CPA.γ_C P (CPA.Ana P) ⊆ CPA.γ_C P (τ_Ω P (ΩCFA.Ana P))] *)
  (* begin hide *)
  intros.
  assert (H1 :CPA.γ_C P (CPA.Ana P) ⊆ CPA.γ_C P (τ_Ω P (ΩCFA.Ana P))).
   apply CPA.γ_C_mono.
   apply CPAΩCFA.
  (* end hide *)
  (* The proof follows by [γ_τ_Ω] and transitivity *)
  (* begin hide *)
  eapply leR_trans ; eauto.
  eapply leR_trans with (B:= CPA.γ_C P (τ_Ω P  (ΩCFA.AnaI' P bullet 0 e0))).
  apply CPA.γ_C_mono. apply τ_Ω_mono. apply AnaAnaI'.
  apply γ_τ_Ω.
  (* end hide *)
Qed.

Lemma CPAbeatsooCFA : forall P, CPA.γ_C P (CPA.Ana P) ⊆ ooCFA.γ_C P (ooCFA.Ana P).
Proof.
  (** By transitivity, [CPAbeatsΩCFA] and [Ωbeatsoo] *)
(* begin hide *)
  intros.
  generalize (CPAbeatsΩCFA P).
  generalize (Ωbeatsoo P).
  intros.
  eapply leR_trans ; eauto.
Qed.
(* end hide *)
End CPAΩCFA.

(* Local Variables: *)
(* coding: utf-8 *)
(* coq-prog-name: "/usr/local/bin/coqtop" *)
(* End: *)
 