(**********************************************************************)
(*                                                                    *)
(*                        CPA beats ooCFA                             *)
(*                                                                    *)
(*       Frédéric Besson,      Inria Rennes - Bretagne Atlantique     *)
(*                                                                    *)
(*  Copyright Institut National de Recherche en Informatique et en    *)
(*  Automatique. All rights reserved                                  *)
(*                                                                    *)
(**********************************************************************)
(** Fixpoint theorems : Kleene and Knaster-Tarski *)
Require Import Setoid. 
Require Import Omega.

(** Order over relations *)
Definition leR (A:Type) (R R' : A -> Prop) : Prop :=
  forall x, R x -> R' x.

Notation "A ⊆ B" := (leR _ A B) (at level 80).

(** Equivalence of relations *)

Definition eqR (A:Type) (R R' : A -> Prop) : Prop :=
  forall x , R x <-> R' x.

Notation "A == B" := (eqR _ A B) (at level 80).

(** Definitions of post-fixpoints and least-fixpoints *)

Definition is_pf (A:Type) (F : (A -> Prop) -> (A-> Prop)) (pf : A -> Prop) :=
 (F pf) ⊆ pf.

Definition is_lfp (A:Type) (F : (A -> Prop) -> (A-> Prop)) (fx : A -> Prop) :=
  is_pf _ F fx /\ forall fx', is_pf _ F fx' ->  fx ⊆ fx'.

(* Aux. result about relations *)

Lemma leR_refl : forall (X:Type) (R :X -> Prop), R ⊆ R.
(* begin hide *)
Proof.
  unfold leR. auto.
Qed.
(* end hide *)

Lemma leR_trans : forall (X:Type) (A B C:X -> Prop), A ⊆ B -> B ⊆ C -> A ⊆ C.
(* begin hide *)
Proof.
  unfold leR. intros. apply H0. apply H. auto.
Qed.
(* end hide *)


Lemma leR_eta_l : forall (X:Type) (R R':X->Prop), R ⊆ R' ->  (fun x => R x) ⊆ R'.
(* begin hide *)
Proof.
  unfold leR.
  intros.
  apply H ; auto.
Qed.
(* end hide *)

Lemma leR_eta_r : forall (X:Type) (R R':X->Prop), R ⊆ R' -> R ⊆ (fun x => R' x).
(* begin hide *)
Proof.
  unfold leR.
  intros.
  apply H ; auto.
Qed.
(* end hide *)

(** The leasf-fixpoint operator is defined as the intersections of the post-fixpoints *)
Definition lfpKT (C:Type) (F : (C -> Prop) -> (C-> Prop)): C -> Prop :=
  fun x => forall (X: C -> Prop) , (F X) ⊆ X -> X x.


Definition mono (C:Type) (F: (C -> Prop) -> (C -> Prop)) :=  forall (e e':C-> Prop), e ⊆ e' ->  (F e) ⊆ (F e').
    
(** If [F] is monotonic then [lfpKT f] is indeed a post-fixpoint of [f] *)
Lemma is_pf_lfpKT : forall (C:Type) (F: (C -> Prop) -> (C -> Prop)), mono C F -> is_pf _ F (lfpKT C F).
(* begin hide *)
Proof.
  unfold lfpKT, is_pf,  leR in *.
  intros until 1. rename H into Fmono.
  intros.
  apply H0.
  eapply Fmono with (2:= H).
  intro. intros.
  apply H1.
  intros.
  apply H0 ; auto.
  revert H2.
  eapply Fmono ; eauto.
  apply leR_eta_l.
  apply leR_refl.
Qed.
(* end hide *)
    
(** [lfpKT f] is the least-fixpoint of [f] *)
Lemma is_lpf_lfpKT : forall (C:Type) (F: (C-> Prop) -> (C -> Prop)), mono C F -> is_lfp _ F (lfpKT C F).
(* begin hide *)
Proof.
  unfold is_lfp.
  split.
  apply is_pf_lfpKT ; auto.
  intros.
  unfold lfpKT.
  intro. intros.
  unfold is_pf in *.
  apply H1. apply leR_eta_r.
  unfold leR. intros. apply H0.
  revert H2.
  apply H.
  apply leR_eta_l.
  apply leR_refl.
Qed.
(* end hide *)

(** The *induction* principle of the fixpoint operator *)
Lemma lfpKT_ind : forall (C:Type) (F: (C-> Prop) -> (C-> Prop)) (X:C->Prop),  mono C F -> (F X) ⊆ X -> lfpKT C F ⊆ X.
(* begin hide *)
Proof.
  unfold lfpKT,leR in *.
  intros until 1.
  rename H into Fmono. intros.
  apply H0. intros. apply H.
  revert H1. apply Fmono. 
  apply leR_eta_l. apply leR_refl.
Qed.
(* end hide *)
  


(** Of course, the least-fixpoint is a fixpoint *)
Lemma lfpKT_fix : forall (C:Type) (F: (C -> Prop) -> (C -> Prop)), mono C F -> lfpKT C F ⊆  F (lfpKT C F).
(* begin hide *)
Proof.
  intros.
  apply lfpKT_ind ; auto.
  apply H.
  apply is_pf_lfpKT ; auto.
Qed.
(* end hide *)

(** The least-fixpoint is *also* the limit of the iterates of the function *)
Fixpoint iter (A:Type) (F : (A -> Prop) -> (A -> Prop)) (n:nat)   :  A -> Prop :=
  match n with
    | O => fun x => False
    | S n => fun x => F (iter A F n ) x
  end.

Definition lub (A:Type) (Rel : nat -> A -> Prop) : (A -> Prop) :=
  fun x => exists n, Rel n x.

Definition lfpK  (C:Type) (F : (C -> Prop) -> (C-> Prop)): C -> Prop := 
  lub C (iter C F).

Definition chain (A:Type) (f : nat -> A -> Prop) :=
  forall n n', n <= n' -> f n ⊆ f n'.

Definition cont (C:Type) (F : (C-> Prop) -> (C -> Prop)):=  
  forall f : nat -> (C -> Prop), 
    chain C f ->  
    (F (lub _ f)) ⊆ (lub _  (fun n => F (f n))).

Lemma cont_morph : forall (C:Type) (F F': (C-> Prop) -> (C -> Prop)),
  (forall x, F x ⊆ F' x)  -> (forall x, F' x ⊆ F x) -> 
  cont C F -> cont C F'.
(* begin hide *)
Proof.
  intros.
  unfold cont in *.
  repeat intro.
  apply H0 in H3.
  apply H1 in H3 ; auto.
  destruct H3. exists x0. apply H ; auto.
Qed.
(* end hide *)  


(** The iterates of a monotonic function form a chain *)
Lemma iter_chain : forall (C:Type) (F : (C-> Prop) -> (C -> Prop)),  mono C F -> chain C (iter C F).
(* begin hide *)
Proof.
  unfold chain. unfold leR.
  intros until n.
  induction n ; simpl ; intros.
  tauto.
  destruct n'.
  apply False_ind ; omega.
  simpl.
  eapply H ; eauto.
  unfold leR ; apply IHn. omega.
Qed.
(* end hide *)
  
(** The least upper-bound of the iterates of F is a post-fixpoint *)
Lemma lfpK_pf : forall (C:Type) (F: (C -> Prop) -> (C -> Prop)), mono C F -> cont C F -> is_pf _ F (lfpK _ F).
(* begin hide *)
Proof.
  unfold is_pf.
  intros.
  unfold cont in H0.
  generalize (H0  _ (iter_chain _ F H)).
  unfold leR,lfpK.
  intros.
  apply H1 in H2.
  unfold lub.
  destruct H2.
  exists (S x0).
  auto.
Qed.
(* end hide*)
    
(* Actually, the least upper-bound of the iterates of F is also the least-fixpoint *)
Lemma is_lpf_lfpK : forall (C:Type) (F: (C -> Prop) -> (C -> Prop)), mono C F -> cont C F -> is_lfp _ F (lfpK _ F).
(* begin hide *)
Proof.
  unfold is_lfp.
  intros.
  generalize (lfpK_pf _ F H H0).
  intros ; split ; auto.
  intros.
  unfold is_pf in *.
  unfold leR in *.
  intros.
  destruct H3.
  revert x H3 .
  induction x0.
  simpl ; tauto.
  simpl.
  intros.
  apply H2.
  eapply H ; eauto.
Qed.
(* end hide *)

Lemma lub_mono : forall (C:Type) (f f': nat -> (C -> Prop)), (forall n, f n ⊆ f' n) -> lub _ f ⊆ lub _ f'.
Proof.
  intros.
  intro. intros. destruct H0. exists x0. eapply H ; eauto.
Qed.

Lemma lfpK_ind : forall (C:Type) (F: (C-> Prop) -> (C-> Prop)) (X:C->Prop),  mono C F -> cont C F -> (F X) ⊆ X -> lfpK C F ⊆ X.
(* begin hide *)
Proof.
  unfold lfpK,leR in *.
  intros until 1.
  rename H into Fmono. intros.
  apply H0. eapply is_lpf_lfpK ; eauto.
  unfold is_pf.
  apply Fmono.  unfold leR. apply H0.
Qed.
(* end hide *)

Lemma lfpK_fix : forall (C:Type) (F: (C -> Prop) -> (C -> Prop)), mono C F -> cont C F -> lfpK C F ⊆  F (lfpK C F).
(* begin hide *)
Proof.
  intros.
  unfold cont in H0.
  unfold leR, lfpK.
  intros.
  destruct H1.
  revert H1. revert x.
  destruct x0.
  simpl. tauto.
  simpl ; intros.
  eapply H ; eauto.
  unfold leR.
  intros. exists  x0. auto.
Qed.
(* end hide *)

Lemma lfpK_mono : forall  (C:Type) (F F': (C -> Prop) -> (C -> Prop)), mono C F -> (forall x, F x ⊆ F' x) ->  lfpK C F ⊆  lfpK C F'.
(* begin hide *)
Proof.
  intros.
  intro. intros.
  destruct H1. exists x0.
  revert x H1.
  induction x0. 
  simpl in *. auto.
  simpl in *.
  intros.
  eapply H0 ; eauto.
  eapply H ; eauto.
Qed.

(** A few results on *staged* fixpoint computations — imbricated fixpoints *)

Definition mono2 (A:Type) (f : (A -> Prop) -> (A -> Prop) -> (A -> Prop)) : Prop :=
  forall x x', x ⊆ x' -> 
    forall y y', y ⊆ y' ->
      f x y ⊆ f x' y'.

Lemma mono_curry : forall (A:Type) (F : (A -> Prop) -> (A -> Prop) -> (A -> Prop)),
  (forall x, mono _ (F x)) -> 
  (forall x, mono _ (fun y => F y x)) -> mono2 A F.
(* begin hide *)
Proof.
  unfold mono2, mono ; intros.
  intro. intros.
  apply H with (e':= y') in H3 ; auto.
  eapply H0 ; eauto.
Qed.
(* end hide *)
  
Require Import Max.

Lemma iter_mono : forall (A:Type) (F F': (A -> Prop) -> (A -> Prop)),
  mono A F' ->
  (forall x, F x ⊆ F' x) -> 
  forall n,
  iter A F n ⊆ iter A F' n.
(* begin hide *)
Proof.
  intros until 0. intros Fmono Fcmp.
  induction n ; simpl.
  intro ; auto.
  intro. intros.
  apply Fcmp in H.
  eapply Fmono ; eauto.
Qed.
(* end hide *)

Lemma iter_mono_n : forall (A:Type) (F : (A -> Prop) -> (A -> Prop)),
  mono A F ->
  forall n n', n <= n' -> 
  iter A F n ⊆ iter A F n'.
(* begin hide *)
Proof.
  intros until 0. intros Fmono.
  induction n ; simpl.
  intros ; intro ; tauto.
  intro. intros.
  destruct n'. apply False_ind ; omega.
  intro. simpl.
  apply Fmono. apply IHn ; omega.
Qed.
(* end hide *)

  
Lemma lfpK_cont : forall (A: Type)  (F : (A -> Prop) -> (A -> Prop) -> (A -> Prop)),
  (forall x, mono _ (F x)) -> 
  (forall x, mono _ (fun y => F y x)) -> 
  (forall y, cont _ (fun x => F x y)) ->
  (forall y, cont _ (fun x => F y x)) ->
  cont _ (fun x => lfpK _ (fun y => F x y)).
(* begin hide *)
Proof.
  unfold cont. intros until F. intros Fmono Fmono' Fcont Fcont' f Hchain.
  apply lfpK_ind.
  (* mono *)
  unfold mono.
  intros. apply Fmono ; auto.
  (* cont *)
  unfold cont; intros.
  apply  Fcont' ; auto.
  (* ind case *)
  intro ; intros.
  apply Fcont in H ; auto.
  destruct H as [n H].
  apply Fcont' in H. destruct H as [n' H].
  unfold lfpK in H. 
  apply Fcont' in H.
  destruct H as [i H].
  exists (max n n').
  exists (S i). simpl.
  revert H.
  eapply mono_curry ; eauto.
  apply Hchain. auto with arith.
  apply iter_mono.
  unfold mono. intros. apply Fmono ;auto.
  intros. apply Fmono'. apply Hchain. auto with arith.
  (* chain iter *)
  unfold chain.
  apply iter_mono_n ; auto.
  unfold mono. intros. apply Fmono ; auto.
  (* chain lfpk *)
  unfold chain.
  intros.
  apply lfpK_mono ; auto.
  unfold mono. intros. apply Fmono ; auto.
  intros.
  apply Fmono'. apply Hchain ; auto with arith.
Qed.
(* end hide *)

Lemma cont2 : forall (C:Type) (F : (C -> Prop) -> (C -> Prop) -> (C -> Prop)),
  (forall x, mono _ (F x)) -> 
  (forall x, mono _ (fun y => F y x)) -> 
  (forall y, cont C (fun x : C -> Prop => F x y)) -> 
  (forall y, cont C (fun x : C -> Prop => F y x)) -> 
  cont _ (fun x => F x x).
(* begin hide *)
Proof.
  intros until F.
  intros Fmono Fmono' Fcont Fcont'.
  unfold cont.
  repeat intro.
  apply Fcont in H0 ; auto.
  destruct H0. 
  apply Fcont' in H0 ; auto.
  destruct H0.
  exists (max x0 x1).
  revert H0.
  eapply mono_curry ; eauto.
  apply H. auto with arith.
  apply H. auto with arith.
Qed.
(* end hide *)



Definition lfpK2 (C : Type) (F : (C -> Prop) -> (C -> Prop) -> (C -> Prop)) : C -> Prop :=
  lfpK _ (fun x1 => lfpK _ (fun x2 => F x1 x2)).


Lemma lfpK2lfpK : forall (C : Type) (F : (C -> Prop) -> (C -> Prop) -> (C -> Prop)), 
  (forall y, mono C (fun x =>  F y x)) -> 
  (forall y, mono C (fun x =>  F x y)) -> 
  (forall y, cont _ (fun x => F x y)) ->
  (forall y, cont _ (fun x => F y x)) ->
  lfpK2 _ F  ⊆ lfpK C (fun x => F x x).
(* begin hide *)
Proof.
  intros until F. intros Fmono1 Fmono2 Fcont1 Fcont2.
  unfold lfpK2.
  eapply lfpK_ind; auto.
  unfold mono ; intros.
  eapply lfpK_mono ; eauto. 
  intro.
  eapply Fmono2 ; eauto.
  apply lfpK_cont ; auto.
  eapply lfpK_ind; auto.
  set (Fx := (lfpK C (fun x : C -> Prop => F x x))).
  change (F Fx Fx) with ((fun x => F x x) Fx). 
  apply lfpK_pf.
  unfold mono.
  intros.
  eapply mono_curry ; eauto.
  eapply cont2 ; eauto.
Qed.
(* end hide *)

Lemma μFusion :  forall (C:Type) (f g h:  (C -> Prop) -> (C -> Prop)),
  mono _ f -> cont _ f -> mono _ h -> cont _ h ->
  mono _ g ->
  (forall x, f (g x) ⊆ g (h x)) -> 
  (lfpK _ f ⊆ g (lfpK _ h)).
(* begin hide *)
Proof.
  intros until h.
  intros monoF contF monoH contH monoG H.
  assert (HH := H (lfpK _ h)).
  assert (HH':f (g (lfpK C h)) ⊆ g (lfpK C h)).
  repeat intro.
  apply HH in H0.
  revert H0.
  apply monoG ; auto.
  apply lfpK_pf ;auto.
  clear HH.
  apply lfpK_ind ; auto.
Qed.
(* end hide *)  


Lemma lfpKlfpK2 : forall (C : Type) (F : (C -> Prop) -> (C -> Prop) -> (C -> Prop)), 
  (forall y, mono C (fun x =>  F y x)) -> 
  (forall y, mono C (fun x =>  F x y)) -> 
  (forall y, cont _ (fun x => F x y)) ->
  (forall y, cont _ (fun x => F y x)) ->
  lfpK C (fun x => F x x) ⊆  lfpK2 _ F.
(* begin hide *)
Proof.
  intros.
  apply lfpK_ind.
  unfold mono.
  intros.
  apply mono_curry ; eauto.
  apply cont2 ; auto.
  repeat intro.
  apply lfpK_pf ; auto.
  unfold mono ; intros.
  apply lfpK_mono;auto.
  intros.
  apply H0;auto.
  apply  lfpK_cont ; auto.
  apply lfpK_pf ; auto.
  revert H3.
  eapply mono_curry ; eauto.
  unfold lfpK2. apply leR_refl.
  change (lfpK C
             (fun x1 : C -> Prop => lfpK C (fun x3 : C -> Prop => F x1 x3))) with (lfpK2 C F) .
  unfold lfpK2.
  set (f := (fun x1 : C -> Prop => lfpK C (fun x2 : C -> Prop => F x1 x2))).
  repeat intro.
  change ((fun x => lfpK C (fun x2 : C -> Prop => F x x2)) (lfpK _ f) x0).
  eapply μFusion with (f:=f) ; unfold f;auto.
  unfold mono. intros. apply lfpK_mono ; auto.
  intro. apply H0 ; auto.
  apply lfpK_cont ;auto.
  unfold mono. intros. apply lfpK_mono ; auto.
  intro. apply H0 ; auto.
  apply lfpK_cont ;auto.  
  unfold mono. intros. apply lfpK_mono ; auto.
  intro. apply H0 ; auto.
  intro.
  apply leR_refl.
Qed.
(* end hide *)

Section AIKT.
  (** We are now able to state basic results for abstract interpretation *)

  Record DomKT (D:Type) : Type := {
    F : (D -> Prop ) -> (D -> Prop) ;
    Fmono : mono D F ; 
      Fcont : cont D F
  }.
  
  (** Consider a concrete domain [C] *)
  Hypothesis C : Type.
  Hypothesis CDom : DomKT C.
  
  (** and an abstract domain [A] *)
  Hypothesis A : Type.
  Hypothesis ADom : DomKT A.
  
  (** linked by a monotonic concretisation function *)
  Hypothesis Gamma : (A -> Prop) -> (C -> Prop).
  Hypothesis Gamma_mono : forall F F', F  ⊆ F' ->  (Gamma F) ⊆ (Gamma F').  
  
  (** such that a one-step abstract computation correctly models a one-step concrete step *)
  Definition Fc := (F C CDom).
  Definition Fa := (F A ADom).
  
  Hypothesis F_gamma : forall e,  (Fc (Gamma e)) ⊆ (Gamma (Fa e)).
  
  (** then, least-fixpoint  commutes *)
  
  Lemma lfp_corrct : (lfpK C Fc) ⊆ (Gamma (lfpK A Fa)).
  (* begin hide *)
  Proof.
    apply lfpK_ind.
    apply Fmono.
    apply Fcont.
    intros.
    unfold leR.
    intros.
    assert (Gamma (Fa (lfpK A Fa)) x).
    eapply F_gamma ; eauto.
    eapply Gamma_mono ; eauto.
    eapply lfpK_pf ; eauto.
    apply Fmono ; eauto.
    apply Fcont.
  Qed.
  (* end hide *)

  (** Sometimes, we have a weaker property *)
  Hypothesis F_gamma_fx :  (Fc (Gamma (lfpK A Fa))) ⊆ (Gamma (Fa (lfpK A Fa))).
  
  Lemma lfp_corrct_fx :  (lfpK C Fc) ⊆ (Gamma (lfpK A Fa)).
  (* begin hide *)
  Proof.
    apply lfpK_ind.
    apply Fmono. apply Fcont.
    intros.
    unfold leR.
    intros.
    assert (Gamma (Fa (lfpK A Fa)) x).
    eapply F_gamma_fx ; eauto.
    eapply Gamma_mono ; eauto.
    eapply lfpK_pf ; eauto.
    apply Fmono ; eauto.
    apply Fcont.
  Qed.
  (* end hide *)
  

End AIKT.


(* Local Variables: *)
(* coding: utf-8 *)
(* coq-prog-name: "/usr/local/bin/coqtop" *)
(* End: *)