(**********************************************************************)
(*                                                                    *)
(*                        CPA beats ooCFA                             *)
(*                                                                    *)
(*       Frédéric Besson,      Inria Rennes - Bretagne Atlantique     *)
(*                                                                    *)
(*  Copyright Institut National de Recherche en Informatique et en    *)
(*  Automatique. All rights reserved                                  *)
(*                                                                    *)
(**********************************************************************)
(** Context-sensitive instrumentation of the semantics *)
Require Import Utf8.
Require Import StdNotations.
Require Import Cool.
Require Import List.

Definition lift_opt (X Y : Type) (f : X -> option Y) (x:option X) : option Y  := 
  match x with None => None | Some v => f v end.

(* begin hide *)
Implicit Arguments lift_opt [X Y].
(* end hide *)

Notation "X ** F" := (lift_opt F X) (at level 80).

(** A context-sensitive analysis is parametrised by:
   - [CC] the set of call-contexts
   - [AC] the set of allocation contexts
   - ∅ the initial call-context
   - [push] for modeling a method call
   - [alloc] for modeling the creation of a new address *)

Module Type Context.
  Parameter CC : Set.
  Parameter bullet : CC. 
  Parameter  AC : Set.
  (** For [alloc] the class parameter is redundant. Likewise, for the method parameter of [push].
     This information could be (painfully) recovered from the program counter. *)
  Parameter  alloc : CC -> PC -> Class -> (Var -> option AC) -> AC.
  Parameter  push : CC -> PC -> Meth   -> (Var -> option AC) -> CC.
  Notation "∅" := bullet.

End Context.

Module ISemantics(Ctx : Context).
  Import Ctx.
  
  (** The instrumented semantics carries along the computation an address cache [Acache] mapping addresses to their abstraction *)
  Definition Acache := Addr -> option AC.
  (** Frames are instrumented by the call context *)
  Record  IFrame :Set := MkF {
    pc  : nat ;
    env : Env ;
    cc  : CC}.
  (** States  are instrumented with an address cache *)
  Record IState : Set := MkS {
    heap : Heap ;
    stack : list IFrame ;
    acache : Addr -> option AC
  }.
  (** We write [IState] and [IFrame] as triples i.e., [ a , b , c ]   *)
  (* begin hide *)
  (* Provide IState and IFrame the [[ A , B , C ]] syntax *)
  Instance IFrameT : Triple nat Env CC IFrame := { cons := MkF}.
  Instance IStateT : Triple Heap (list IFrame) (Addr -> option AC) IState := { cons := MkS}.
  (* end hide *)

  (** In order to call the [alloc] and [push] context function, 
     the instrumented semantics need to apply the adress cache [ac] to an environment [e]. *)

  (** In the paper, [abs e ac] is written e^ac *)
  Definition abs (e: Env) (ac:Acache) : Var -> option AC :=
    fun v =>  e v ** ac.

   (** Except for [New], intra-procedural statements do not modify the instrumentation.
      - The [New] statement updates the address cache
      - The [Call] statement updates the call-context
      - The [Ret] statement restores the call-context *)
  Inductive isos (P:Program) : IState -> IState -> Prop :=
  | isos_intra : forall pc i e h e' h' s cc ac, nthp P pc = Some i -> ~ (exists c,  i  = New c) -> intra i (e,h) (e',h') -> 
    isos P ([ h , ( [ pc , e ,cc ]::s) , ac ]) ([ h' ,  ( [ pc+1 , e' , cc]::s) , ac ])
  | isos_alloc : forall pc  e h  s cc ac c a, nthp P pc = Some (New c) -> 
    h a = None -> 
    isos P
    ([ h , ([pc, e,cc]::s), ac])
    ([ h [ a ↦  Some ( c, Null) ] , ([pc+1 , e [ v0 ↦ Some a] , cc] ::s) , ac [ a ↦ Some (alloc cc pc c (abs e ac))]])
  | isos_call : forall pc m e h s ac cc, 
    nthp P pc = Some (Call m) -> 
    isos P
    ([ h, [pc,e,cc]::s, ac])
    ([ h, [lk P (dp h e,m),e, push cc pc m (abs e ac)]::[pc,e,cc] ::s, ac])
  | isos_ret : forall (pc pc':nat) h (e e': Env) (s:list IFrame) (cc cc':CC) (ac:Acache), 
    nthp P pc = Some Ret -> 
    isos P
    ([ h, ([pc,e,cc]::[pc',e',cc']::s) , ac])
    ([ h,  ([pc'+1, e' [v0  ↦  e(v0)] , cc']::s) ,ac]).

  (** The initial instrumented state [is0] *)
  Definition h0 : Heap := fun a => None.
  Definition ac0 : Acache := fun a => None.
  Definition e0 : Env := fun v => None.

  Definition is0 : IState := [h0 , [ 0 , e0 , ∅ ] ::nil , ac0].

  (** Instrumented reachable states [Iacc]. *)
  Inductive IAcc (P:Program) : IState -> Prop :=
  | IAcc0 : IAcc P is0
  | IAccIsos : forall s s', IAcc P s ->  isos P s s' -> IAcc P s'. 

  (** An alternative definition - to get the (right) induction principle over the length of the derivation *)
  Inductive IAccN (P:Program) : nat -> IState -> Prop :=
    IAccNO : IAccN P 0 is0
  | IAccNS : forall n s s', IAccN P n s ->  isos P s s' -> IAccN P (Datatypes.S n) s'. 

  Lemma IAccIAccN : forall P is, IAcc P is -> exists n, IAccN P n is.
  (* begin hide *)
  Proof.
    intros.
    induction H. exists 0. constructor.
    destruct IHIAcc. exists (Datatypes.S x).
    eapply IAccNS ; eauto.
  Qed.
  (* end hide *)

  Lemma IAccNIAcc : forall P n is, IAccN P n is -> IAcc P is.
  (* begin hide *)
  Proof.
    intros.
    induction H. constructor.
    eapply IAccIsos ; eauto.
  Qed.
  (* end hide *)

  (** To recover a non-instrumented state [State] from an instrumented
     state [IState] , we define an erasure function [EE] which traverses the istate and removes the instrumentation.
     *)
  Definition eraseFrame (f : IFrame) : Frame :=
    let (pc,e,cc) := f in (pc, e).
  
  Definition eraseStack (sf : list IFrame) : list Frame := List.map eraseFrame sf.

  Definition EE : IState -> State   := fun is => let (h,s,ac) := is in (h,eraseStack s).

  Lemma new_or_not_new : forall (i:St) , (exists c, i = New c) \/ ~ (exists c, i = New c).
  (* begin hide *)
  Proof. destruct i ; try (right ; intro H ; destruct H ; discriminate). left. exists c. reflexivity.
  Qed. 
  (* end hide *)
     
  (** The instrumentation is not intrusive. 
      - Lemma [AccIAcc] shows that there all the reachable states are captured
      - Lemma [IAccAcc] shows that the instrumentation does not enable new states *)
   Lemma AccIAcc : forall P s, Acc P s -> exists is, IAcc P is /\ s = EE is.
   Proof.
   (** The proof is by induction over the definition of [Acc] 
      and case analysis of the definition of [sos].
      The proof follows because is a 1-1 mapping between the [sos] and [isos] rules *)
   (* begin hide *)
     intros.
     induction H.
     (* base case *)
     exists is0. split ; [constructor | reflexivity].
     (* inductive case *)
     destruct IHAcc as [is [Hacc Heq]].
     destruct is as [h sf ac].
     (* case analysis over sos *)
     inv  H0.
     (* intra rules *)
     simpl in H3. destruct sf ; [discriminate | inv H3].
     destruct i0 ; inv H5.
     destruct (new_or_not_new i).
     (* in case this is a new, the address cache is updated *)
     destruct H0. subst. inv H2.
     exists (MkS (h [a ↦ Some (x, Null)])
       ((MkF (pc1 + 1) (env0 [v0 ↦ Some a]) cc0) ::  sf) (ac [ a ↦ Some (alloc cc0 pc1 x (abs env0 ac)) ])).
     split. econstructor 2 ; eauto.
     constructor ; auto. reflexivity.
     (* this is not a new *)
     exists (MkS h' ((MkF (pc1+1) e' cc0) ::sf) ac).
     split.
     econstructor ; eauto.
     eapply isos_intra ; eauto. 
     reflexivity.
     (* call *)
     destruct sf ; simpl in H2 ; [discriminate | inv H2].
     destruct i ; simpl in H4 ; inv H4.
     exists (MkS h ((MkF (Cool.lk P (dp h env0, m)) env0 (push cc0 pc1 m (abs env0 ac))) :: (MkF pc1 env0 cc0) ::sf) ac).
     split. econstructor ; eauto. eapply isos_call ; eauto.
     reflexivity.
     (* return *)
     destruct sf ; [discriminate |inv H2].
     destruct sf ; [discriminate |inv H5].
     exists (MkS h (MkF (pc' + 1) (e'[v0 ↦ (e v0) ]) (cc i0) :: sf) ac).
     split. econstructor ; eauto.
     destruct i0 ; destruct i ; simpl in *.
     inv H4 ; inv H2.
     eapply isos_ret ; eauto.
     reflexivity.
    (* end hide *)
   Qed.

   Lemma IAccAcc : forall P is, IAcc P is -> Acc P (EE is).
   Proof.
   (** The proof is by induction over the definition of [IAcc] and case analysis over  [isos]. *)
   (* begin hide *)
     intros.
     induction H.
     constructor.
     econstructor;eauto.
     inv H0. simpl. eapply sos_intra ; eauto.
     simpl. eapply sos_intra ; eauto.
     constructor ; auto.
     simpl. constructor ; auto.
     simpl. constructor ; auto.
   (* end hide *)
   Qed.

   (** Semantic invariants *)
   (** The soundness proof the context-sensitive control-flow analysis depends on several semantic invariants that we detail now:
      - Absence of dangling pointers: All the addresses in the stack frame are bound in the heap and all the
      fields of the objects in the heap refer themselves to addresses
      allocated in the heap (or possibly the null pointer).
      *)

   Definition wf_heap (h:Heap) : Prop := 
     (forall a t o, h a = Some (t, o)  -> forall fd a', o fd = Some a' -> h a' <> None).

   Definition wf_env (e:Env) (h:Heap): Prop  :=
     forall r a, e r = Some a -> h a <> None.

   Definition wf_H (st : IState) : Prop := 
     forall (ac:Acache)  (h:Heap) (s:list IFrame), (st = [h , s , ac ]) -> 
       wf_heap h
       /\
       (forall (pc:PC) (e:Env) cc, (In ([pc , e, cc]) s) -> wf_env e h).
 
   (** The initial state [is0] is trivially well-formed. *)
   Lemma wf_H_is0 : wf_H is0.
   (* begin hide *)
   Proof.
     unfold is0.
     unfold wf_H ; intros. inv H.
     split. unfold wf_heap. intros.  discriminate.
     unfold wf_env ; intros. simpl in H. destruct H ; auto. inv H. discriminate.
   Qed.
   (* end hide *)

   Lemma IAcc_wf_H : forall P is, IAcc P is -> wf_H is.
   Proof.
     (** The proof is by induction over the defintition of [IAcc] *)
     (* begin hide *)
     intros. unfold wf_H, wf_heap, wf_env.
     induction H.
     (* base case *)
     apply wf_H_is0.
     (* IH *)
     inv H0.
     inv H3.
     (* copy *)
     destruct (IHIAcc _ _ _ (refl_equal ([h', ([pc0, e, cc0]) :: s0, ac]))) as [HH1 HH2].
     split; inv H0 ; intros.
     eapply HH1 ; eauto.
     simpl in H0.
     destruct H0.
     inv  H0.
     subst_h r r0  H3.
     eapply HH2 ; eauto. left ; reflexivity.
     eapply HH2 ; eauto. left ; reflexivity.
     eapply HH2 ; eauto. right ; apply H0.
     (* read *) 
     destruct (IHIAcc _ _ _ (refl_equal ([h', ([pc0, e, cc0]) :: s0, ac]))) as [HH1 HH2].
     split; inv H0 ; intros.
     eapply HH1 ; eauto.
     simpl in H0.
     destruct H0.
     inv  H0.
     subst_h v0 r  H3.
     eapply HH1 ; eauto. 
     eapply HH2 ; eauto. left ; reflexivity.
     eapply HH2 ; eauto. right ; apply H0.
     (* write *) 
     destruct (IHIAcc _ _ _ (refl_equal ([h, ([pc0, e', cc0]) :: s0, ac]))) as [HH1 HH2].
     split; inv H0 ; intros.
     subst_g a a'. congruence.
     subst_h a a0 H0. inv H0.
     subst_h f fd H3. inv H3.
     eapply HH2 ; eauto. left ; reflexivity.
     eapply HH1 ; eauto. 
     eapply HH1 ; eauto. 
     subst_g a a0. congruence.
     simpl in H0.
     destruct H0.
     inv  H0.
     eapply HH2 ; eauto. left ; reflexivity.
     eapply HH2 ; eauto. right ; apply H0.
     (**)
     apply False_ind.
     apply H2. exists t ; reflexivity.
     (*new*)
     destruct (IHIAcc _ _ _ (refl_equal ([h, ([pc0, e, cc0]) :: s0, ac]))) as [HH1 HH2].
     split; inv H0 ; intros.
     subst_g a a'. congruence.
     subst_h a a0 H0. inv H0. discriminate.
     eapply HH1 ; eauto. 
     subst_g a a0. congruence.
     simpl in H0.
     destruct H0.
     inv  H0.
     subst_h v0 r H3. inv H3. congruence.
     eapply HH2 ; eauto. left ; reflexivity.
     eapply HH2 ; eauto. right ; apply H0.
     (* call *)
     destruct (IHIAcc _ _ _ (refl_equal ([h, ([pc0, e, cc0]) :: s0, ac]))) as [HH1 HH2].   
     split; inv H0 ; intros.   
     eapply HH1 ; eauto.
     simpl in H0. destruct H0.
     inv H0.
     eapply HH2 ; eauto. left ; reflexivity.
     destruct H0. inv H0. eapply HH2 ; eauto.  left ; reflexivity.
     eapply HH2 ; eauto. right. apply H0.
     (* return *) 
     destruct (IHIAcc _ _ _ (refl_equal ([h, ([pc0, e, cc0]) :: ([pc', e', cc']) :: s0, ac]))) as [HH1 HH2].       
     split; inv H0 ; intros.      
     eapply HH1 ; eauto.
     simpl in H0 ; destruct H0.
     inv H0.
     subst_h v0 r H2.
     eapply HH2 ; eauto. left ; reflexivity.
     eapply HH2 ; eauto. right. left ; reflexivity.
     eapply HH2 ; eauto. right. right. apply H0.
     (* end hide *)
   Qed.
 
   (** We define an ordering [le_ac] over adress caches.
      Basically, [le_ac ac ac'] holds if [ac(a) = ac'(a)] or [ac(a)] is not defined. *)
   Definition le_ac (ac ac' : Acache) : Prop :=
     forall a ab, ac a = Some ab -> ac' a = Some ab.
   
   (** The relation [le_ac] is reflexive [le_ac_refl] and transitive [le_ac_trans]. *)
   (* begin hide *)
   Lemma le_ac_refl : forall x, le_ac x x.
   Proof.
     unfold le_ac.
     auto.
   Qed.
   
   Lemma le_ac_trans : forall x y z,
     le_ac x y -> le_ac y z -> le_ac x z.
   Proof.
     unfold le_ac ; intros.
     apply H0. apply H. auto.
   Qed.
   (* end hide *)


   (** We proof that an address is defined in the heap only and only if it is defined in the adress cache *)
   Lemma IAccAc : forall P is, IAcc P is -> forall a, heap is a = None <-> acache is a = None.
   (* begin hide *)
   Proof.
     intros P is Hacc.
     induction Hacc.
     simpl. unfold h0. unfold ac0. tauto.
     inv H ; simpl in *.
     inv H2;auto.
     (* write *)
     intros.
     subst_g a a0.
     generalize (IHHacc a0). intuition congruence.
     (* new *)
     apply False_ind. apply H1. exists t ; reflexivity.
     intros.
     subst_g a a0.
     subst_g a0 a0 ; try discriminate.
     intuition congruence. congruence.
     subst_g a a0 ; try congruence.
     auto.
     auto.
   Qed.
   (* end hide *)


   (** We prove that the frames of a semantic state follow a stack discipline.
      Roughly speaking, if we a state [[h,f0::f1::s, ac]] is reachable
      then 
      - a state [[h',f1::s,ac']] was *previously* reachable such that
      - the program counter of f1 is a call
      - the call context of f0 is obtained by computing push from the context of [f1]
      - there are more addresses in ac than in ac'
      *)
   Lemma IAccNret : forall P n (h:Heap) (pc0:PC) (e:Env) cc0 pc' e' cc' s1 ac,
     IAccN P n ( [ h, ([pc0, e, cc0]) :: ([pc', e', cc']) :: s1, ac]) -> 
     exists n', n' < n /\
     exists h', exists m,  
       nthp P pc' = Some (Call m) /\ 
       exists ac', le_ac ac' ac /\
         push cc' pc' m (abs e' ac') = cc0 /\
         IAccN P n' ([h', ([pc', e', cc']) :: s1, ac']).
   Proof.
     (** The proof is by *strong* induction over [n] *)
     (** intra-procedural cases are solved by the induction hypothesis at step n-1 *)
     (** the call case is trivially solved *)
     (** the handle the [ret] case, the induction hypothesis is used twice to unwind the stack *)
     (* begin hide *)
     induction n using Wf_nat.lt_wf_ind.
     intros.
     inv H0.
     assert (Hlt : n0 < Datatypes.S n0) by omega.
     inv H2.
     (* intra *)
     destruct (H _ Hlt _ _ _ _ _ _ _ _ _ H1).
     exists x.
     split. omega. tauto.
     (* new *)
     destruct (H _ Hlt _ _ _ _ _ _ _ _ _ H1).
     exists x.
     split. omega. 
     destruct H0.
     destruct H2 as [h' [m [Hnth [ac' [Hle [Hpush Hacc]]]]]].
     exists h'. exists m. split ; auto. exists ac'.
     intuition.
     eapply le_ac_trans ; eauto.
     unfold le_ac.
     intros. subst_g a a0.
     apply False_ind.
     apply IAccNIAcc in H1. 
     apply IAccAc with (a:= a0) in H1.
     simpl in H1. intuition congruence.
     simpl ; auto.
     (* call *)
     exists n0.
     split. omega.
     exists h;exists m.
     split ; auto.
     exists ac. split.
     apply le_ac_refl.
     tauto.
     (* ret *)
     destruct (H _ Hlt _ _ _ _ _ _ _ _ _ H1).
     destruct H0.
     destruct H2 as [h' [m [Hnth [ac' [Hle [Hpush Hacc]]]]]].
     assert (x < Datatypes.S n0) by omega.
     destruct (H _ H2 _ _ _ _ _ _ _ _ _ Hacc).
     destruct H3 as [Hlt2 [h2 [m2 [Hnth2 [ac2 [Hle2 [Hpush2 Hacc2]]]]]]].
     exists x0.
     split. omega.
     exists h2. exists m2. split ; auto.
     exists ac2. split.
     eapply le_ac_trans ; eauto.
     tauto.
     (* end hide *)
   Qed.

   


End ISemantics.


(* Local Variables: *)
(* coding: utf-8 *)
(* coq-prog-name: "/usr/local/bin/coqtop" *)
(* End: *)
