(**********************************************************************)
(*                                                                    *)
(*                        CPA beats ooCFA                             *)
(*                                                                    *)
(*       Frédéric Besson,      Inria Rennes - Bretagne Atlantique     *)
(*                                                                    *)
(*  Copyright Institut National de Recherche en Informatique et en    *)
(*  Automatique. All rights reserved                                  *)
(*                                                                    *)
(**********************************************************************)
(** Specification of the points-to analysis *)
Require Import Utf8.
Require Import List.
Require Import StdNotations.
Require Import Fix.
Require Import Cool.
Require Import ICool.


Module Analysis(Ctx : Context).
  (** The principle of context-sensitive analyses is to abstract the set of instrumented states [IAcc].
     - addresses are abstracted by allocation contexts [AC]
     - call-stacks by the call contexts [CC]
     - sets of environments  [Env] are abstracted in a non-relational manner (see Andersen's analysis)
     *)
  Export Ctx.

  Module ISem := ISemantics(Ctx).
  Export ISem.


  (** The analysis is defined over the following abstract domains **)
  Definition HClass := AC -> Class -> Prop.
  Definition HObject := AC -> Field -> option AC -> Prop.
  Definition AEnv := Var -> option AC -> Prop.
  Definition CCEnv := CC -> PC -> AEnv.

  (** Because the analysis essentially computes relations, we could model the result
     of the analysis as three mutually inductive relations (HC,HO,E) ∈ HClass * HObject * CCEnv
   *)
  (** However, this is not a wise choice. In Coq, dealing with mutually recursive relations is technically feasible but complicated. *)
  (** Instead, as it is customary in abstract interpretation, the result of the analysis is expressed as the least fixpoint of a functional. *)
  (** In the following, we define three functional
     - [HClassSpec] defines [HC] w.r.t  [E]
     - [HObjectSpec] defines [HO] w.r.t [HC] and [E] 
     - [EnvSpec] defines [E]  w.r.t [HC] and [HO] and [E]
     
     Eventually, we define [ASpecF] the global functional of the analysis *)

   (** To define these objects, two auxiliary definitions are needed *)

  (** For [e : Var -> option AC] and [pe : Var -> option AC -> Prop] ,
     We write [e ∈ pe] iff [∀  v, e v ∈ pe(v)] *)
  Definition mem (e: Var -> option AC) (pe : Var -> option AC -> Prop) :=
    forall v, pe v (e v).
  Notation "A ∈ B" := (mem A B) (at level 90).



  (** The dispatch function [dp] is abstracted by the [adp] relation defined below *)
  Definition adp (HC: HClass) (e: Var -> option AC) (d : Var -> Class) : Prop :=
    forall v , match e v with
                 | None => d v = Udf
                 | Some a => HC a (d v)
               end.

  (** [HClassSpec] is the simplest relation. Its role is to dynamically bind an allocation context [a] to its class [c].
     For most analysis, this binding needs not being computed because the class [c] is usually a function of [a].
     For instance, when [AC] is the allocation point. Yet, in full generality, this could be a dynamic property. *)

  Inductive HClassSpec (P:Program) (E :CCEnv) : HClass :=
  | HTS_alloc : forall pc cc   t es, nthp P pc = Some (New t) -> 
    es ∈ (E cc pc)  -> 
    HClassSpec P E (alloc cc pc t es) t.

  (** [HObjectSpec] is a points-to graph. Each side-effect [write] statement adds edges in the abstract heap.
     At allocation time, we also record that the fields are [null] – this could be left implicit at the cost of complicating the concretisation function *)
  Inductive HObjectSpec (P:Program)  (HT : HClass) (E : CCEnv) : HObject :=
  | HOS_alloc : forall pc f t cc es, nthp P pc = Some (New t) -> 
    es ∈  (E cc pc)  ->
    HObjectSpec P HT E (alloc cc pc t es) f None
  | HOS_write : forall pc f cc e, nthp P pc = Some (Write f) -> e ∈  (E cc pc) -> 
    forall a b, e v0 = (Some a) -> e v1 = b -> 
      HObjectSpec P HT E a f b.

  Section EnvSpec.

  (** Instead of [EnvSpec], we define [EnvSpecI] a generalisation of [EnvSpec]  parametrised by :
      - an initial context [cci]
      - an initial program counter [pci]
      - and an initial abstract environment [ei]
      *)
   (** We will need [EnvSpecI] to state intermediate lemmas and compare CPA and ∞-CFA.
       [EnvSpec] is [EnvSpecI] with cci=∅ , pci = 0 and ei = λ x . null *)

    Hypothesis P:Program.
    Hypothesis cci : CC.
    Hypothesis pci : PC.
    Hypothesis ei  : Var -> option AC.

    Inductive EnvSpecI  (HT : HClass) (HO : HObject)  (EE : CCEnv) : CCEnv :=
    | ESI_init :  forall r, EnvSpecI HT HO EE cci pci r (ei r)
    | ESI_call : forall cc  m pc t (e:Var -> option AC), 
      nthp P pc = Some (Call m) -> 
      e ∈ EE cc pc -> 
      adp HT e t ->  
      forall r,  EnvSpecI HT HO EE (push cc pc m e) (lk P (t,m)) r (e r)
    | ESI_ret0 : forall pc pc' m  cc'  e er,
      nthp P pc = Some Ret -> 
      nthp P pc' = Some (Call m) -> 
      er ∈  (EE (push cc' pc' m e) pc) -> 
      e ∈  (EE cc' pc') -> 
      EnvSpecI  HT HO EE cc' (pc'+1) v0 (er v0)
    | ESI_ret' : forall  pc' m r cc'  er, 
      nthp P pc' = Some (Call m) -> 
      er ∈ (EE cc' pc') -> 
      r <> v0 -> 
      EnvSpecI HT HO EE cc' (pc'+1) r (er r)
    | intra_read_r0 : forall cc pc a f b e,
      nthp P pc = Some (Read f) -> 
      e ∈  (EE cc pc) -> 
      e v0 = (Some a) -> 
      HO a f b -> 
      EnvSpecI HT HO EE cc (pc+1) v0 b
    | intra_read_r' : forall cc pc r f e,
      nthp P pc = Some (Read f) -> 
      e ∈ (EE cc pc) -> 
      EnvSpecI HT HO  EE cc (pc+1)  r (e r)
    | intra_copy_r0 : forall cc pc r r' e,
      nthp P pc = Some (Copy r r') -> 
      e ∈  (EE cc pc) ->
      EnvSpecI  HT HO EE cc (pc+1)  r (e r')
    | intra_copy_r' : forall cc pc r r' r'' e,
      nthp P pc = Some (Copy r r') -> r'' <> r -> 
      e ∈  (EE cc pc) -> 
      EnvSpecI  HT HO EE cc (pc+1)  r'' (e r'')
    | intra_write : forall  cc pc f r e,
      nthp P pc = Some (Write f) -> 
      e ∈ (EE cc pc) -> 
      EnvSpecI  HT HO EE cc (pc+1)  r (e r)
    | intra_alloc_r0 : forall cc pc t  es, 
      nthp P pc = Some (New t) -> 
      es ∈  (EE cc pc) -> 
      EnvSpecI HT HO EE cc (pc+1)  v0 (Some (alloc cc pc t es))
    | intra_alloc_r' : forall cc pc r t  es, 
      nthp P pc = Some (New t) -> 
      es ∈ (EE cc pc) -> 
      r <> v0 -> 
      EnvSpecI  HT HO EE cc (pc+1)  r (es r).
  
    (** The type [Atom] is a sum type combining the elements of the different relations computed by the analysis *)
    (** Hence, the result of the context-sensitive points-to analysis is a set of atoms : [Atom -> Prop] *)
    Inductive Atom := 
    | HC : AC ->  Class -> Atom
    | HO : AC -> Field -> option AC -> Atom
    | Ecc : CC -> PC -> Var -> option AC -> Atom.

    Definition AIState := Atom -> Prop.

    (** The following aux. functions filter atoms according to the constructor *)
    Definition gHC (S : AIState) : HClass := fun a t => S (HC a t).
    
    Definition gHO (S: AIState) : HObject := fun a f b => S (HO a f b).
    
    Definition gEnv (S:AIState) : CCEnv :=  fun cc pc r a => S (Ecc cc pc r a).
    
    Definition mkA (Ht : HClass) (Ho : HObject) (E : CCEnv) : AIState :=
      fun a => 
        match a with 
          | HC a t => Ht a t
          | HO a f a' => Ho a f a'
          | Ecc c cp r a => E c cp r a
        end.


    (** Given these definitions, we can eventually give the functional [ASpecFI] of the points-to analysis *)
    Inductive ASpecFI  (AS :AIState) : AIState :=
    | AnaObjectF : forall a f b, HObjectSpec P (gHC AS) (gEnv AS) a f b -> 
      ASpecFI  AS (HO a f b)
    | AnaTagF : forall a t,
      HClassSpec P (gEnv AS) a t -> ASpecFI  AS  (HC a t)
    | AnaEnvF : forall cc pc r a,
      EnvSpecI  (gHC AS) (gHO AS) (gEnv AS) cc pc r a -> 
      ASpecFI AS (Ecc cc pc r a).
  
    (** [AnaI] is the result of the analysis starting from  initial conditions [cci] [pci] [ei] **)
    Definition AnaI := lfpK _ ASpecFI.

    (** For proving "CPA beats ooCFA" this specification is not totally adequate.
       Therefore, we provide an equivalent definition [AnaI'] which in essence computes the fixpoint in a *staged* fashion. *)
    Inductive CCEnvT :=
      Ecc' : CC -> PC -> Var -> option AC -> CCEnvT.

    Definition mkCCEnv (e : CCEnvT -> Prop) : CCEnv :=
      fun cc pc r a => e (Ecc' cc pc r a).

    Definition destCCEnv (e: CCEnv) : CCEnvT -> Prop :=
      fun t => let (cc,pc,r,a):= t in e cc pc r a.

    Definition EnvSpecF (HC : HClass) (HO:HObject) (E: CCEnvT -> Prop) : (CCEnvT -> Prop) :=
      destCCEnv (EnvSpecI HC HO (mkCCEnv E)).

    Definition EnvSpecIFx (HC : HClass) (HO:HObject) := mkCCEnv (lfpK _ (EnvSpecF HC HO)).

    Definition ASpecCurry (A : AIState) (A' : AIState) : AIState :=
      ASpecFI (mkA (gHC A) (gHO A) (gEnv A')).

    Definition ASpecFI' (A : AIState) : AIState :=
      ASpecCurry A (mkA (gHC A) (gHO A) (EnvSpecIFx (gHC A) (gHO A))).

    Definition AnaI' := lfpK _ ASpecFI'.

    (** To reason about these fixpoint we prove that the functional are monotone [mono] and continuous [cont].
       Proving monotony is only annoying but not difficult; proving continuity is usually less intuitive.
       Both properties are yet very compositional. 
       *)
    (* begin hide *)
    Definition leEnv (e e': CCEnv) : Prop :=
      forall cc pc r a, e cc pc r a -> e' cc pc r a.

    Lemma leEnv_gEnv_mono : forall e e' , e ⊆ e' ->  leEnv (gEnv e) (gEnv e').
    Proof.
      intros.
      unfold gEnv in *. unfold leEnv.
      intros.
      eapply H ; eauto.
    Qed.

    Lemma leEnv_in_mono : forall e e' cc pc es, leEnv e e' -> es ∈ e cc pc -> es ∈  e' cc pc.
    Proof.
      intros.
      unfold mem in *.  
      intros. eapply H ; eauto.
    Qed.

    Lemma in_gEnv_mono : forall e e' cc pc es, e ⊆ e' -> es ∈ gEnv e cc pc -> es ∈ gEnv e' cc pc.
    Proof.
      intros.
      eapply leEnv_in_mono ; eauto.
      eapply leEnv_gEnv_mono ; eauto.
    Qed.


    Definition leHC (ht : HClass) (ht' : HClass) : Prop :=
      forall a t, ht a t -> ht' a t.

    Lemma gHC_mono : forall e e', e ⊆ e' -> leHC  (gHC e) (gHC e').
    Proof.
      unfold gHC, leHC.
      intros. eapply H ; eauto.
    Qed.

    Definition leHO (ho : HObject) (ho' : HObject) : Prop :=
      forall a f b, ho a f b-> ho' a f b.


    Lemma leHC_refl : forall ht, leHC ht ht.
    Proof.
      unfold leHC ; auto.
    Qed.

    Lemma leHO_refl : forall ho, leHO ho ho.
    Proof.
      unfold leHO ; auto.
    Qed.

    Lemma mkA_mono : forall hc hc' ho ho' e e',
      leHC hc hc' ->  leHO ho ho' -> leEnv e e' -> 
      mkA hc ho e ⊆ mkA hc' ho' e'.
    Proof.
      unfold mkA.
      unfold leR ; intros.
      destruct x.
      eapply H; eauto.
      eapply H0 ; eauto.
      eapply H1 ; eauto.
    Qed.


    Lemma adp_mono : forall hc hc', leHC hc hc' -> forall a c, 
      adp hc a c -> adp hc' a c.
    Proof.
      unfold adp.
      intros.
      generalize (H0 v).
      destruct (a v). eapply H ; eauto.
      auto.
    Qed.

    Lemma adp_gHC_mono : forall e e', e ⊆ e' -> forall a c, 
      adp (gHC e) a c -> adp (gHC e') a c.
    Proof.
      intros. eapply adp_mono ; eauto. eapply gHC_mono ; eauto.
    Qed.

    Lemma gHO_mono : forall e e', e ⊆ e' -> leHO (gHO e) (gHO e').
    Proof.
      unfold gHO , leHO.
      intros. eapply H ; eauto.
    Qed.

    Lemma EnvSpec_mono : forall hc hc' ho ho' e e', 
      leHC hc hc' -> leHO ho ho' -> leEnv e e' -> 
      leEnv (EnvSpecI hc ho e) (EnvSpecI hc' ho' e').
    Proof.
      intros.
      unfold leEnv. intros.
      inv H2. 
      rewrite <- H3.
      rewrite <- H4.
      rewrite <- H5.
      constructor.
    (* call *)
      constructor; auto. eapply leEnv_in_mono ; eauto.
      eapply adp_mono ; eauto.
    (* ret *)
      eapply ESI_ret0 with (e:=e1);eauto.
      eapply leEnv_in_mono ; eauto.
      eapply leEnv_in_mono; eauto.
    (* ret' *)
      eapply ESI_ret' ; eauto.
      eapply leEnv_in_mono ; eauto.
    (* intra *)
      eapply intra_read_r0 with (a:= a0); eauto.
      eapply leEnv_in_mono ; eauto.
    (**)
      eapply intra_read_r' ; eauto.
      eapply leEnv_in_mono ; eauto.
    (**)
      eapply intra_copy_r0 ; eauto.
      eapply leEnv_in_mono ; eauto.
    (**)
      eapply intra_copy_r' ; eauto.
      eapply leEnv_in_mono ; eauto.
    (**)
      eapply intra_write ; eauto.
      eapply leEnv_in_mono ; eauto.
    (**)
      eapply intra_alloc_r0 ; eauto.
      eapply leEnv_in_mono ; eauto.
      eapply intra_alloc_r' ; eauto.
      eapply leEnv_in_mono ; eauto.
    Qed.

    Lemma HObjectSpec_mono : forall hc hc' e e',
      leHC hc hc' -> 
      leEnv e e' -> 
      leHO (HObjectSpec P hc e) (HObjectSpec P hc' e').
    Proof.
      intros.
      unfold leHO.
      intros. inv H1. constructor. auto.
      eapply leEnv_in_mono ; eauto.
      eapply HOS_write with (cc:= cc0) ; eauto.
      eapply leEnv_in_mono ; eauto.
    Qed.

    Lemma HClassSpec_mono : forall e e',
      leEnv e e' -> 
      leHC (HClassSpec P e) (HClassSpec P e').
    Proof.
      intros.
      unfold leHC.
      intros. inv H0.
      constructor ; auto.
      eapply leEnv_in_mono ; eauto.
    Qed.

    Lemma  ASpecF_mono :  mono Atom ASpecFI.
    Proof.
      unfold mono.
      intros.
      intro. intros.
      inv H0.
      constructor.
    (* HObjectSpec *)
      eapply HObjectSpec_mono ; eauto.
      apply gHC_mono ; auto.
      apply leEnv_gEnv_mono ; auto.
    (* HClassSpec *)
      constructor.
      eapply HClassSpec_mono ; eauto.
      apply leEnv_gEnv_mono ; auto.
    (* EnvSpec *)
      constructor. 
      eapply EnvSpec_mono ; eauto.
      eapply gHC_mono ; eauto.
      eapply gHO_mono ; eauto.
      eapply leEnv_gEnv_mono ; eauto.
    Qed.


    Require Import FType.

    Lemma mem_cont : forall es f cc0 pc0, 
      chain Atom f -> 
      es ∈ gEnv (lub Atom f) cc0 pc0 -> 
      exists n,  es ∈ (gEnv (f n) cc0 pc0).
    Proof.
      unfold mem, lub.
      intros.
      unfold gEnv in *.
      set (PP := (fun n v => f n ((Ecc cc0 pc0 v (es v))))).
      change (exists n, forall v, PP n v).
      apply ex_swap with (AllT := Vars) (P:= PP).
      exact v0.
      apply finVar. 
      apply H0.
      unfold PP.  unfold chain in H.
      intros.
      unfold leR. intro. apply H ; auto.
    Qed.

    Lemma adp_cont : forall e f t, 
      chain _ f -> 
      adp (gHC (lub Atom f)) e t -> 
      exists n, adp (gHC (f n)) e t.
    Proof.
      intros.
      unfold adp in *.
      unfold lub in H0.
      assert (forall v, exists n, match e v with
                                    | Some a =>  gHC (f n) a (t v)
                                    | None  => t v = Udf
                                  end).
       intros.
       generalize (H0 v). destruct (e v).
       unfold gHC. auto. intro.  econstructor ; eauto.
      clear H0.
      eapply ex_swap in H1. destruct H1. 
      exists x. intros. generalize (H0 v).
      destruct (e v);auto.
      exact v0.
      apply finVar.
      unfold leR. intros.
      destruct (e x).  unfold gHC in *. eapply H ; eauto.
      auto.
    Qed.

    Lemma gEnv_cont : forall f c p v o, 
      lub Atom f (Ecc c p v o) -> 
      gEnv (lub Atom f) c p v o.
    Proof.
      intros.
      destruct H.
      unfold gEnv. exists x ; auto.
    Qed.

    Lemma gHO_cont : forall f a fd b,
      chain _ f -> 
      gHO (lub Atom f) a fd b -> 
      exists n, gHO (f n) a fd b.
    Proof.
      unfold lub, gHO.
      intros. destruct H0. exists x.
      auto.
    Qed.

    Lemma gHC_cont : forall f a  b,
      chain _ f -> 
      gHC (lub Atom f) a  b -> 
      exists n, gHC (f n) a  b.
    Proof.
      unfold lub, gHC.
      intros. destruct H0. exists x.
      auto.
    Qed.

    Require Import Max.
    Ltac prove_cont :=
      match goal with
        | H : ?E ∈ gEnv (lub Atom ?F) ?CC ?PC |- _ => apply (mem_cont E F CC PC) in H ; auto ; let n := fresh "n" in destruct H as [n H]
        | H : adp (gHC (lub Atom ?F)) ?E ?X |- _ => apply adp_cont in H ; auto ; let n := fresh "n" in destruct H as [n H]
        | H : gHO (lub Atom ?F) ?A ?FD ?B |- _  => apply gHO_cont in H ; auto ; let n := fresh "n" in destruct H as [n H]
      end.

    Ltac prove_mono :=
      repeat 
        (match goal with
           |  H : ?E ∈ ?EE ?CC ?PC |- ?E' ∈ gEnv ?EE' ?CC ?PC => eapply leEnv_in_mono ; eauto
           |     |- leEnv (gEnv ?E) (gEnv ?E') => eapply leEnv_gEnv_mono ; eauto
           | H : adp (gHC ?E) ?e ?t |- adp (gHC ?E') ?e ?t => eapply adp_mono with (2:= H) ; eauto
           | |- leHC (gHC ?E) (gHC ?E') => apply gHC_mono ; auto
         end).


    Lemma mem_gEnv_mkA : forall er HC HO E cc pc,
      er ∈ gEnv (mkA HC HO E) cc pc <->  er ∈ E cc pc.
    Proof.
      unfold gEnv, mkA,mem.
      tauto.
    Qed.

    Lemma adp_gHC_mkA : forall A B C e t,
      adp (gHC (mkA A B C)) e t <->  adp A e t.
    Proof.
      unfold gHC, mkA.
      unfold adp. intros. 
      tauto.
    Qed.

    Lemma gHO_mkA : forall A B C a b c,
      gHO (mkA A B C) a b c <-> 
      B a b c.
    Proof.
      unfold gHO, mkA.
      tauto.
    Qed.



    Lemma mem_mkCCEnv_mono : forall er E E' cc pc,
      E ⊆ E' -> 
      er ∈ mkCCEnv E cc pc ->
      er ∈ mkCCEnv E' cc pc.
    Proof.
      unfold mem, mkCCEnv.
      intros.  eapply H ; eauto.
    Qed.

    Ltac simpl_mkA := 
      match goal with
        | H :?E ∈ gEnv (mkA ?A ?B ?C) ?CC ?PC |- _ => 
          rewrite mem_gEnv_mkA  in H
        | H : ?E ∈ mkCCEnv ?E'' ?CC ?PC |- ?E ∈ mkCCEnv ?E' ?CC ?PC => eapply mem_mkCCEnv_mono with (2:=H); eauto
        | H : adp (gHC (mkA ?A ?B ?C)) ?E ?T |- _ => 
          rewrite (adp_gHC_mkA A B C E T) in H
        | H : gHO (mkA ?A ?B ?C) ?a ?b ?c |- _ =>
          rewrite gHO_mkA in H
        |    |- ?E ∈ gEnv (mkA ?A ?B ?C) ?cc ?pc => rewrite mem_gEnv_mkA
      end.

    Lemma ASpecF_cont1 : forall x, cont Atom (fun a => ASpecCurry a x).

    Proof.
      unfold cont ,  leR ; intros.
      unfold ASpecCurry in H0.
      inv H0.
      (* HObj *)
      inv H1 ; simpl_mkA.
      exists 1. constructor. constructor ; auto.
      exists 1. constructor. eapply HOS_write ; eauto.
      (* HClass *)
      inv H1 ; simpl_mkA.
      exists 1. constructor. constructor ; auto.
      (* EnvSpec *)
      inv H1 ; repeat simpl_mkA.
      (* *)
      exists 0. constructor. rewrite <- H0. rewrite <- H2. constructor.
      (* call *)
      repeat prove_cont. 
      exists n.
      constructor. eapply ESI_call ; eauto.
      (* call / ret *)
      exists 1.
      constructor. eapply ESI_ret0 with (e:=e); eauto.
      (* ret' *)
      exists 1. constructor. eapply ESI_ret' ; eauto.
      (* read *)
      prove_cont.
      exists n.
      constructor. eapply intra_read_r0; eauto.
      (* read *)
      exists 1. constructor. eapply intra_read_r'; eauto.
      (* copy *)
      exists 1. constructor. eapply intra_copy_r0; eauto.
      exists 1. constructor. eapply intra_copy_r'; eauto.
      (* write *)
      exists 1. constructor. eapply intra_write; eauto.
      (* alloc *)
      exists 1. constructor. eapply intra_alloc_r0; eauto.
      exists 1. constructor. eapply intra_alloc_r'; eauto.
    Qed.

    Lemma ASpecF_cont2 : forall x, cont Atom (fun a => ASpecCurry x a).
    Proof.
      unfold cont ,  leR ; intros.
      unfold ASpecCurry in H0.
      inv H0.
      (* HObj *)
      inv H1 ; simpl_mkA ; prove_cont.
      exists n. constructor. constructor ; auto.
      exists n. constructor. eapply HOS_write ; eauto.
      (* HClass *)
      inv H1 ; simpl_mkA ; prove_cont.
      exists n. constructor. constructor ; auto.
      (* EnvSpec *)
      inv H1 ; repeat simpl_mkA ; repeat prove_cont.
      (* *)
      exists 0. constructor. rewrite <- H0. rewrite <- H2. constructor.
      (* call *)
      exists n.
      constructor. eapply ESI_call ; eauto.
      (* call / ret *)
      exists (max n n0).
      constructor. eapply ESI_ret0 with (e:=e); eauto ; simpl_mkA.
      prove_mono. prove_mono.
      apply H. auto with arith.
      prove_mono. prove_mono.
      apply H. auto with arith.
      (* ret' *)
      exists n. constructor. eapply ESI_ret' ; eauto.
      (* read *)
      exists n.
      constructor. eapply intra_read_r0; eauto.
      (* read *)
      exists n. constructor. eapply intra_read_r'; eauto.
      (* copy *)
      exists n. constructor. eapply intra_copy_r0; eauto.
      exists n. constructor. eapply intra_copy_r'; eauto.
      (* write *)
      exists n. constructor. eapply intra_write; eauto.
      (* alloc *)
      exists n. constructor. eapply intra_alloc_r0; eauto.
      exists n. constructor. eapply intra_alloc_r'; eauto.
    Qed.

    Lemma ASpecCurry_mono1 : forall x, mono _ (fun y => ASpecCurry y x).
    Proof.
      unfold mono, ASpecCurry ; intro. 
      intros.
      apply ASpecF_mono. apply mkA_mono. 
      apply gHC_mono ; auto.
      apply gHO_mono ; auto.
      unfold leEnv ; auto. 
    Qed.

    Lemma ASpecCurry_mono2 : forall x, mono _ (fun y => ASpecCurry x y).
    Proof.
      unfold mono, ASpecCurry ; intro. 
      intros.
      apply ASpecF_mono. apply mkA_mono. 
      unfold leHC ; auto.
      unfold leHO ; auto.
      apply leEnv_gEnv_mono ; auto.
    Qed.
      
    Lemma ASpecF_cont : cont Atom ASpecFI.
    Proof.
      assert (cont _ (fun x => ASpecCurry x x)).
      generalize ASpecCurry_mono1.
      generalize ASpecCurry_mono2.
      generalize ASpecF_cont1.
      generalize ASpecF_cont2.
      intros.
      apply cont2 ; auto.
      unfold cont in *.
      repeat intro.
      assert (ASpecCurry (lub _ f) (lub _ f) x).
      unfold ASpecCurry.  revert H1. apply ASpecF_mono ; auto.
      intro.
      intros.
      destruct x0 ; simpl.
      eapply gHC_cont ; eauto.
      eapply gHO_cont ; eauto.
      eapply gEnv_cont ; eauto.
      apply H in H2 ; auto.
      destruct H2. exists x0.
      unfold ASpecCurry in H2.
      revert H2.
      apply ASpecF_mono ; auto.
      unfold mkA. repeat intro.
      destruct x1 ; auto.
    Qed.

    Lemma AnaI_pf : ASpecFI AnaI ⊆ AnaI.
    Proof.
      eapply lfpK_pf ; eauto.
      apply ASpecF_mono.
      apply ASpecF_cont.
    Qed.

    Lemma AnaI_fix :  AnaI ⊆ ASpecFI AnaI.
    Proof.
      eapply lfpK_fix ; eauto.
      apply ASpecF_mono.
      apply ASpecF_cont.
    Qed.



    Lemma EnvSpecF_mono :  forall hc hc' ho ho' e e',
      leHC hc hc' -> leHO ho ho' -> e ⊆ e' -> 
      EnvSpecF hc ho e ⊆ EnvSpecF hc' ho' e'.
    Proof.
      unfold EnvSpecF. intros.
      intro.
      unfold destCCEnv. destruct x.
      eapply EnvSpec_mono ; eauto.
      unfold mkCCEnv. unfold leEnv.
      intros; eapply H1 ; eauto.
    Qed.

    Lemma EnvSpec_mkCCEnv : forall HC HO HC' HO' f c p v o,
      EnvSpecI HC HO (mkCCEnv (lub CCEnvT f)) c p v o -> 
      EnvSpecI HC HO (gEnv (lub Atom (fun n =>  mkA HC' HO' (mkCCEnv (f n))))) c p v o.
    Proof.
      intros.
      eapply EnvSpec_mono ; eauto.
      unfold leHC ; auto.
      unfold leHO ; auto.
      unfold leEnv, mkCCEnv.
      intros. unfold gEnv. destruct H0. exists x.
      unfold mkA. auto.
    Qed.

    Lemma EnvSpecF_cont : forall HC HO, cont _ (EnvSpecF HC HO).
    Proof.
      unfold cont ; intros.
      intro.
      intros.
      unfold EnvSpecF in H0.
      unfold destCCEnv in H0.
      destruct x.
      apply EnvSpec_mkCCEnv with (HC' := HC0) (HO' := HO0) in H0 .
      assert (Hchain :chain Atom (fun n : nat => mkA HC0 HO0 (mkCCEnv (f n)))).
       unfold chain in * ; intros.
       unfold leR. intro. unfold mkA. destruct x ;auto.
       unfold mkCCEnv. apply H ; auto.
      unfold EnvSpecF, destCCEnv.
      inv H0.
      exists 0. rewrite <- H1. rewrite <- H2. constructor.
      (* call *)
      repeat prove_cont. 
      exists n.
      eapply ESI_call ; eauto.
      (* call / ret *)
      repeat prove_cont.
      exists (max n n0).
      eapply ESI_ret0 with (e:=e); eauto.
      repeat simpl_mkA.
      apply H. auto with arith.
      repeat simpl_mkA.
      apply H. auto with arith.
      (* ret' *)
      prove_cont.
      exists n. eapply ESI_ret' ; eauto.
      (* read *)
      repeat prove_cont. 
      exists n.
      eapply intra_read_r0; eauto.
      (* read *)
      prove_cont.
      exists n.  eapply intra_read_r'; eauto.
      (* copy *)
      prove_cont.
      exists n. eapply intra_copy_r0; eauto.
      prove_cont.
      exists n. eapply intra_copy_r'; eauto.
      (* write *)
      prove_cont.
      exists n. eapply intra_write; eauto.
      (* alloc *)
      prove_cont.
      exists n. eapply intra_alloc_r0; eauto.
      prove_cont. 
      exists n. eapply intra_alloc_r'; eauto.
    Qed.

    Definition AnaICurry := lfpK2 _ ASpecCurry.

    Lemma AnaIAnaICurry : AnaI ⊆ AnaICurry.
    Proof.
      unfold AnaI, AnaICurry.
      repeat intro.
      apply lfpKlfpK2.
      apply ASpecCurry_mono2 ; auto.
      apply ASpecCurry_mono1 ; auto.
      apply ASpecF_cont1 ; auto.
      apply ASpecF_cont2 ; auto.
      revert H.
      apply lfpK_mono ; auto.
      apply ASpecF_mono.
      intros. unfold ASpecCurry.
      apply ASpecF_mono.
      intro xx ; destruct xx ; auto.
    Qed.

    Lemma AnaICurryAnaI : AnaICurry ⊆ AnaI.
    (* begin hide *)
    Proof.
      unfold AnaI, AnaICurry.
      repeat intro.
      apply lfpK2lfpK in H.
      revert H.
      apply lfpK_mono ; auto.
      unfold mono ; intros.
      unfold ASpecCurry.
      apply ASpecF_mono.
      apply mkA_mono ; auto.
      apply gHC_mono ; auto.
      apply gHO_mono ; auto.
      apply leEnv_gEnv_mono ; auto.
      intro.
      unfold ASpecCurry.
      apply ASpecF_mono.
      intro xx ; destruct xx ; auto.
      apply ASpecCurry_mono2 ; auto.
      apply ASpecCurry_mono1 ; auto.
      apply ASpecF_cont1 ; auto.
      apply ASpecF_cont2 ; auto.
    Qed.


    Lemma ASpecFI'_ASpecCurry : forall x,
      lfpK _ (fun y => ASpecCurry x y) ⊆ ASpecFI' x.
    Proof.
      intros.
      repeat intro.
      revert H.
      apply lfpK_ind.
      apply ASpecCurry_mono2.
      apply ASpecF_cont2.
      unfold ASpecFI', ASpecCurry. 
      apply ASpecF_mono ; auto.
      apply mkA_mono ; auto.
      unfold leHC ; auto.
      unfold leHO; auto.
      unfold leEnv, gEnv.
      intros. inv H.
      unfold mkA.
      unfold EnvSpecIFx. unfold mkCCEnv.
      apply lfpK_pf ; auto.
      unfold mono ; intros. apply EnvSpecF_mono ; auto.
      unfold leHC ; auto.
      unfold leHO ; auto.
      apply EnvSpecF_cont.
    Qed.

    Lemma ASpecCurry_ASpecFI' : forall x,
      ASpecFI' x  ⊆ lfpK _ (fun y => ASpecCurry x y).
    Proof.
      intros.
      unfold ASpecFI'. unfold ASpecCurry.
      repeat intro.
      apply lfpK_pf.
      unfold mono ; intros.
      eapply ASpecF_mono ; eauto.
      eapply mkA_mono.
      unfold leHC ; auto.
      unfold leHO ; auto.
      apply leEnv_gEnv_mono ; auto.
      apply ASpecF_cont2.
      revert H.
      apply ASpecF_mono.
      apply mkA_mono.
      unfold leHC ; auto.
      unfold leHO ; auto.
      unfold leEnv.
      intros.
      unfold gEnv, mkA in H.
      unfold gEnv.
      unfold EnvSpecIFx in H.
      unfold mkCCEnv in H.
      set (FF:=  lfpK Atom
     (fun y : Atom → Prop =>
      ASpecFI
        (mkA (gHC x) (gHO x)
           (fun (cc1 : CC) (pc1 : PC) (r0 : Var) (a0 : option AC) =>
            y (Ecc cc1 pc1 r0 a0))))).
      change ((fun (x: CCEnvT) => let (cc0,pc0,r,a) := x in FF (Ecc cc0 pc0 r a)) (Ecc' cc0 pc0 r a)).
      eapply lfpK_ind ; eauto.
      unfold mono. intros. apply EnvSpecF_mono ; auto.
      unfold leHC ; auto.
      unfold leHO ; auto.
      apply EnvSpecF_cont.
      repeat intro.
      destruct x1.
      unfold FF.
      apply lfpK_pf ; auto.
      unfold mono ; intros.
      apply ASpecF_mono ; auto.
      apply mkA_mono ; auto.
      unfold leHC ;auto.
      unfold leHO ;auto.
      unfold leEnv ; auto.
      (**)
      apply ASpecF_cont2.
      constructor.
      unfold EnvSpecF in H0.
      unfold destCCEnv in H0.
      revert H0.
      unfold FF. unfold mkCCEnv .
      apply EnvSpec_mono.
      unfold leHC, gHC,mkA.
      auto.
      unfold leHO, gHO, mkA.  auto.
      unfold leEnv, gEnv.
      intros. unfold mkA. apply H0.
    Qed.

    Lemma ASpecFI'_mono : mono _ ASpecFI'.
    Proof.
      unfold mono ;   intros.
      repeat intro.
      apply ASpecCurry_ASpecFI' in H0.
      apply ASpecFI'_ASpecCurry.
      revert H0. apply lfpK_mono.
      apply ASpecCurry_mono2.
      intros. apply ASpecCurry_mono1 ; auto.
    Qed.

    Lemma ASpecFI'_cont : cont _ ASpecFI'.
    Proof.
      intros.
      apply cont_morph with (F:= fun x => lfpK _ (fun y => ASpecCurry x y)).
      apply ASpecFI'_ASpecCurry.
      apply ASpecCurry_ASpecFI'.
      apply lfpK_cont.
      apply ASpecCurry_mono2.
      apply ASpecCurry_mono1.
      apply ASpecF_cont1.
      apply ASpecF_cont2.
    Qed.

  
    Lemma AnaICurryAnaI' : AnaICurry ⊆ AnaI'.
    Proof.
      unfold AnaICurry, AnaI'. 
      unfold ASpecCurry.
      unfold lfpK2. 
      apply lfpK_mono.
      unfold mono ; intros.
      apply lfpK_mono.
      apply ASpecCurry_mono2.
      intro.
      apply ASpecCurry_mono1 ; auto.
      apply ASpecFI'_ASpecCurry.
    Qed.

    Lemma AnaI'AnaICurry : AnaI' ⊆ AnaICurry.
    Proof.
      unfold AnaI', AnaICurry.
      unfold lfpK2.
      apply lfpK_mono.
      apply ASpecFI'_mono.
      intro.
      apply ASpecCurry_ASpecFI'.
    Qed.
  (* end hide *)

  Lemma AnaAnaI' : AnaI  ⊆ AnaI'.
  (* begin hide *)
  Proof.
    eapply leR_trans with (B:= AnaICurry ).
    apply AnaIAnaICurry. apply AnaICurryAnaI'.
  Qed.
  (* end hide *)

  Lemma AnaI'Ana : AnaI' ⊆ AnaI.
  (* begin hide *)
  Proof.
    eapply leR_trans with (B:= AnaICurry).
    apply AnaI'AnaICurry. apply AnaICurryAnaI. 
  Qed.
  (* end hide *)
  
  End EnvSpec.

  (* begin hide *)
  Hint Immediate leHO_refl : mono.

  Hint Immediate leHC_refl : mono.
  (* end hide *)

  Definition e0 : Var -> option AC := fun r => None.


  (** The result of the points-to analysis from initial conditions *)
  Definition Ana  (P:Program) : AIState := AnaI P ∅ 0 e0.
  Definition Ana' (P:Program) : AIState := AnaI' P ∅ 0 e0.

  Definition EnvSpecFx (P:Program) (HC : HClass) (HO:HObject) := EnvSpecIFx P ∅ 0 e0 HC HO.

  Definition ASpecF (P:Program) := ASpecFI P ∅ 0 e0.

  Definition EnvSpec (P:Program) (HC : HClass) (HO:HObject) : CCEnv -> CCEnv :=
    EnvSpecI P ∅ 0 e0 HC HO.

  (** [EnvSpecFx] is the fixpoint of [EnvSpec] *)
  Lemma EnvSpecFx_fix : forall P HC HO, leEnv (EnvSpecFx P HC HO) (EnvSpec P HC HO (EnvSpecFx P HC HO)).
  (* begin hide *)
  Proof.
    unfold leEnv, EnvSpecFx, EnvSpecIFx.
    unfold mkCCEnv.
    intros.
    apply lfpK_fix in H.
    eapply EnvSpec_mono ; eauto with mono.
    unfold leEnv. unfold mkCCEnv. auto.
    unfold mono. intros. 
    apply EnvSpecF_mono; auto with mono.
    apply EnvSpecF_cont.
  Qed.
  (* end hide *)

  Lemma EnvSpecFx_pf : forall P HC HO, leEnv (EnvSpec P HC HO (EnvSpecFx P HC HO)) (EnvSpecFx P HC HO).
  (* begin hide *)
  Proof.
    unfold leEnv, EnvSpecFx, EnvSpecIFx.
    unfold mkCCEnv.
    intros.
    apply lfpK_pf ; auto.
    unfold mono ; intros.
    apply EnvSpecF_mono; auto with mono.
    apply EnvSpecF_cont.
  Qed.
  (* end hide *)

  Lemma gEnvAna'EnvSpecFx : forall P, leEnv (gEnv (Ana' P)) (EnvSpecFx P (gHC (Ana' P)) (gHO (Ana' P))).
  (* begin hide *)
  Proof.
    unfold Ana', AnaI'.
    unfold leEnv,gEnv.
    intros.
    apply lfpK_fix in H.
    inv H.
    eapply EnvSpecFx_pf.
    eapply EnvSpec_mono ; eauto.
    unfold leHC.
    intros.
    unfold gHC in H. simpl in H. auto.
    unfold leHO ; intros.
    unfold gHO in H. simpl in H. auto.
    unfold leEnv ; intros.
    unfold gEnv in H. simpl in H. auto.
    apply ASpecFI'_mono.
    apply ASpecFI'_cont.
  Qed.
  (* end hide *)

  Lemma EnvSpecFxgEnvAna' : forall P, leEnv (EnvSpecFx P (gHC (Ana' P)) (gHO (Ana' P ))) (gEnv (Ana' P)).
  (* begin hide *)
  Proof.
    unfold Ana', AnaI'.
    unfold leEnv,gEnv.
    intros.
    apply lfpK_pf.
    apply ASpecFI'_mono.
    apply ASpecFI'_cont.
    constructor.
    eapply EnvSpecFx_fix.
    eapply lfpK_mono ; eauto.
    unfold mono ; intros.
    apply EnvSpecF_mono ; auto with mono.
    intros.
    apply EnvSpecF_mono ; auto with mono.
    apply leR_refl.
  Qed.
  (* end hide *)

  (** The correctness theorem is stated via concretisation functions *)
  Definition γ_H (ac : Acache) (HC : HClass) (HO : HObject) (h : Heap) : Prop :=
    forall a, match h a , ac a with
                | None , None => True
                | Some (c,o) , Some aa => HC aa c /\ forall f, HO aa f ( (o f) ** ac)
                |   _        ,  _  => False
              end.

  Definition γ_E (ac: Acache) (E : AEnv) (e : Env) : Prop :=
    (abs e ac) ∈ E.


  Inductive γ_F (P:Program) (E:CCEnv) : Acache -> list IFrame -> Prop :=
  | F_init : forall ac e pc, γ_E ac (E ∅ pc) e -> γ_F P E ac  ((MkF pc  e  ∅ ) ::nil)
  | F_cons : forall ac ac' e e' pc' pc l m cc, 
    γ_F P E ac  ( (MkF pc e cc) :: l) -> 
    nthp P pc = Some (Call m) -> le_ac ac ac' ->
    γ_E  ac' (E (push cc pc m (abs e ac)) pc') e' -> 
    γ_F P E ac' ( (MkF pc' e' (push cc pc m (abs e ac))) :: (MkF pc e cc) :: l).

  Inductive γ (P:Program) (S:AIState) : IState -> Prop :=
  | G : forall ac l h (Hf : γ_F P (gEnv S) ac  l) (Hh : γ_H ac (gHC S) (gHO S) h), γ P S (MkS h l ac).

  
  (** The concretisations are monotonic *)
  (* begin hide *)
  Lemma γ_E_mono : forall ac s s' cc pc e,
    leEnv s s' -> 
    γ_E ac (s cc pc) e ->
    γ_E ac (s' cc pc) e.
  Proof.
    intros.
    unfold γ_E in *.
    eapply leEnv_in_mono ; eauto.
  Qed.

  Lemma γ_F_mono : forall P ac s s' l,
    leEnv s s' ->     
    γ_F P s ac  l -> 
    γ_F P s' ac  l.
  Proof.
    intros.
    induction H0. constructor.
    eapply γ_E_mono ;eauto.
    eapply F_cons ; eauto.
    eapply γ_E_mono ;eauto.
  Qed.

  Lemma γ_H_mono : forall s s' ac h,
    s ⊆ s' -> 
    γ_H ac (gHC s) (gHO s) h -> 
    γ_H ac (gHC s') (gHO s') h.
  Proof.
    unfold γ_H in *.
    intros.
    generalize (H0 a) ; clear H0.
    destruct (h a);auto.
    destruct o ; auto.
    destruct (ac a) ;auto.
    intros. destruct H0 ; split.
    eapply gHC_mono ; eauto.
    intros.     eapply gHO_mono ; eauto.
  Qed.
  (* end hide *)

  Lemma γ_mono : forall P s s', s ⊆ s' -> (γ P s ) ⊆ (γ P s' ).
  (* begin hide *)
  Proof.
    intros.
    intro.
    intros. inv H0.
    constructor.
    eapply γ_F_mono ; eauto.
    apply leEnv_gEnv_mono ; auto.
    eapply γ_H_mono ; eauto.
  Qed.
  (* end hide *)

  Lemma γ_F_γ_E : forall P ac E e l pc cc, γ_F P E ac ((MkF pc e cc)::l) -> γ_E ac (E cc pc) e.
  (* begin hide *)
  Proof.
    intros.
    inv H ;auto.
  Qed.

  (** We are now able to state the soundness theorem [soundness : forall P s, IAcc P s -> γ P (Ana P) s] *)

  (* begin hide *)
  Lemma ana_pf : forall P, ASpecF P (Ana P)  ⊆  Ana P.
  Proof.
    unfold Ana.
    intros.
    apply AnaI_pf ; auto.
  Qed.

  Lemma ana_fix : forall P,  Ana P ⊆ ASpecF P (Ana P).
  Proof.
    intros.
    apply AnaI_fix ; auto.
  Qed.

  Lemma γ_morph : forall P e e', e == e' -> γ P e == γ P e'.
  Proof.
    unfold eqR.
    intros.
    split. eapply γ_mono ; eauto.
    intro. rewrite H. auto.
    eapply γ_mono ; eauto.
    intro.  rewrite H. auto.
  Qed.


  Lemma γ_morph_Ana : forall P, γ P (ASpecF P (Ana P)) ⊆ γ P (Ana P).  
  Proof.
    intros.
    apply γ_mono.
    apply ana_pf.
  Qed.

  Lemma γ_morph_Ana' : forall P, γ P (Ana P) ⊆  γ P (ASpecF P (Ana P)) .  
  Proof.
    intros.
    apply γ_mono.
    apply ana_fix.
  Qed.
  (* end hide *)

  Lemma γ_is0 : forall P,  γ P (Ana P) is0.
  Proof.
    (** The proof is by definition of γ *)
    (* begin hide *)
    intros.
    apply γ_morph_Ana.
    constructor.
    constructor.
    unfold γ_E. unfold abs. unfold mem. intros. unfold lift_opt. unfold ISem.e0.
    change (@None AC) with (e0 v).
    constructor. constructor.
    constructor.
    (* end hide *)
  Qed.

  (* begin hide *)
  Lemma γ_E_ASpecF : forall  P ac cc0 pc0 e,  
    γ_E ac (gEnv  (ASpecF P (Ana P)) cc0 pc0) e <-> 
    γ_E ac
    (EnvSpec P (gHC (Ana P)) (gHO (Ana P)) (gEnv (Ana P))cc0 pc0) e.
  Proof.
    unfold γ_E in *.
    intros.
    unfold abs in *.
    unfold mem in *.
    split ; intros ;   assert (HH := H v).
    inv HH. auto.
    constructor. auto.
  Qed.

  Lemma γ_F_ASpecF : forall P ac l, 
    γ_F P (gEnv (Ana P)) ac  l -> 
    γ_F P (gEnv (ASpecF P (Ana P))) ac  l.
  Proof.
    intros.
    eapply γ_F_mono ; eauto.
    apply leEnv_gEnv_mono.
    apply ana_fix.
  Qed.

  Lemma gHC_ASpecF : forall P a t, 
    gHC (Ana P) a t ->  gHC (ASpecF P (Ana P)) a t.
  Proof.
    intros. eapply gHC_mono ; eauto.
    eapply ana_fix ; eauto.
  Qed.

  Lemma gHO_ASpecF : forall P a f b,
    gHO (Ana P) a f b ->
    gHO (ASpecF P (Ana P)) a f b.
  Proof.
    intros.
    eapply gHO_mono ; eauto.
    eapply ana_fix ; eauto.
  Qed.

  (* end hide *)


  Lemma γ_E_abs : forall ac cc0 pc0 e A,
    γ_E ac (gEnv A cc0 pc0) e -> 
    (abs e ac) ∈ (gEnv A cc0 pc0).
  (* begin hide *)
  Proof.
    unfold γ_E,abs,mem.
    intros.
    apply H.
  Qed.
  (* end hide *)

  Lemma γ_E_lift : forall ac cc0 pc0 e A r',
    γ_E ac (gEnv A cc0 pc0) e -> 
   gEnv A cc0 pc0 r' (e r' ** ac).
  (* begin hide *)
  Proof.
    unfold γ_E,abs,mem.
    intros.
    apply H.
  Qed.
  (* end hide *)
  
  Lemma γ_E_copy : forall P  cc pc r r' e ac,
    nthp P pc = Some (Copy r r') -> 
    γ_E ac (gEnv  (Ana P) cc pc) e ->
    γ_E ac (gEnv (ASpecF P (Ana P)) cc (pc + 1)) (e [r ↦ e r']).
  (* begin hide *)
  Proof.
    intros.
    rewrite γ_E_ASpecF.
    unfold γ_E.
    unfold abs, mem.
    intros.
    subst_g r v.
    change (e r' ** ac) with ((abs e ac) r').
    eapply intra_copy_r0 ;eauto.
    change (e v ** ac) with ((abs e ac) v).
    eapply intra_copy_r' ;eauto.
  Qed.
  (* end hide *)

  Lemma γ_E_read : forall P f cc pc e h' a t o ac, 
    nthp P pc = Some (Read f) -> 
    e v0 = Some a -> 
    h' a = Some (t, o) ->
    γ_H ac (gHC (Ana P)) (gHO (Ana P)) h' -> 
    γ_E ac (gEnv (Ana P) cc pc) e -> 
    γ_E ac (gEnv (ASpecF P (Ana P)) cc (pc + 1)) (e [v0 ↦ o f]).
    (* begin hide *)
  Proof.
    intros.
    rewrite γ_E_ASpecF.
    unfold γ_E.
    unfold abs, mem.
    intros.
    subst_g v0 v.
    assert (HH := H2 a).
    rewrite H1 in HH. 
    case_eq  (ac a) ; intros; try congruence.
    rewrite H4 in HH.
    eapply intra_read_r0 with (a:=a0);eauto.
    replace (Some a0) with (e v0 ** ac).
    reflexivity.
    unfold lift_opt;  rewrite H0. auto.
    destruct HH. apply H6 ; auto.
    (* ac a = None *)
    rewrite H4 in HH. tauto.
    change (e v ** ac) with ((abs e ac) v).
    eapply intra_read_r';eauto.
  Qed.
    (* end hide *)

  Lemma γ_E_write : forall P  cc pc f e ac,
    nthp P pc = Some (Write f) -> 
    γ_E ac (gEnv  (Ana P) cc pc) e ->
    γ_E ac (gEnv (ASpecF P (Ana P)) cc (pc + 1)) e.
  Proof.
    (* begin hide *)
    intros.
    rewrite γ_E_ASpecF.
    unfold γ_E.
    unfold abs, mem.
    intros.
    change (e v ** ac) with (abs e ac v).
    eapply intra_write ;eauto.
    (* end hide *)
  Qed.

  Lemma γ_H_write : forall P f ac cc0 pc0 e' a t o h, 
    nthp P pc0 = Some (Write f) ->
    e' v0 = Some a -> 
    h a = Some (t, o) -> 
    γ_E ac (gEnv (Ana P) cc0 pc0) e' -> 
    γ_H ac (gHC (Ana P)) (gHO (Ana P)) h -> 
   γ_H ac (gHC (ASpecF P (Ana P))) (gHO (ASpecF P (Ana P)))
     (h [a ↦ Some (t, o [f ↦ e' v1])]).
  (* begin hide *)
  Proof.
    intros.
    unfold γ_H.
    intros.
    subst_g a a0.
    assert (HH := H3 a0).
    rewrite H1 in HH.
    revert HH.
    case_eq (ac a0) ; intros.
    intuition.
    eapply gHC_ASpecF ; eauto.
    subst_g f f0.
    constructor.
    eapply HOS_write with (cc:=cc0) ; eauto.
    replace (Some a) with (e' v0 ** ac).
    reflexivity.
    unfold lift_opt. rewrite H0. auto.
    (**)
    eapply gHO_ASpecF ; eauto.
    auto.
    generalize (H3 a0).    
    case_eq (h a0).
    case_eq (ac a0).
    intros. destruct p. destruct H7 ; split.
    eapply gHC_ASpecF ; eauto.
    intros;eapply gHO_ASpecF ; eauto.    
    destruct p. auto.
    case_eq (ac a0);auto.
  Qed.
  (* end hide *)

  Lemma γ_E_new : forall P ac pc0 cc0 c e h a
    (Ha :forall v', e v' <> Some a) ,
    nthp P pc0 = Some (New c) -> 
    h a = None -> 
    γ_H ac (gHC (Ana P)) (gHO (Ana P)) h -> 
    γ_E ac (gEnv (Ana P) cc0 pc0) e -> 
    γ_E (ac [a ↦ Some (alloc cc0 pc0 c (abs e ac))])
     (gEnv (ASpecF P (Ana P)) cc0 (pc0 + 1)) (e [v0 ↦ Some a]).
  Proof.
    (* begin hide *)
    intros.
    rewrite γ_E_ASpecF.
    unfold γ_E.
    unfold abs, mem.
    intros.
    subst_g v0 v.
    simpl. rewrite subst_eq.
    eapply intra_alloc_r0 ;eauto.
    match goal with
      | |- context[ e v ** ?AC] => change (e v ** AC) with (abs e AC v)
    end.
    eapply intra_alloc_r' ;eauto.
    (**)
    unfold abs. unfold mem.
    unfold lift_opt.
    intros.
    case_eq (e v0).
    intros.
    subst_g a a0.
    apply False_ind. apply (Ha v0). auto.
    replace (ac a0) with (e v0 ** ac).
    eapply γ_E_lift ; eauto.
    rewrite H4. reflexivity.
    intro.
    replace (@None AC) with (e v0 ** ac).
    eapply γ_E_lift ; eauto.    
    rewrite H4. reflexivity.
  Qed.
  (* end hide *)


  Lemma γ_H_new : forall P ac pc0 cc0 c e h a,
    γ_E ac (gEnv (Ana P) cc0 pc0) e -> 
    nthp P pc0 = Some (New c) -> 
    wf_heap h -> 
    h a = None -> 
    γ_H ac (gHC (Ana P)) (gHO (Ana P)) h -> 
    γ_H (ac [a ↦ Some (alloc cc0 pc0 c (abs e ac))]) 
    (gHC (ASpecF P (Ana P))) (gHO (ASpecF P (Ana P)))
    (h [a ↦ Some (c, Null)]).
(* begin hide *)
  Proof.
    intros.
    unfold γ_H.
    intros.
    subst_g a a0.
    subst_g a0 a0 ; try discriminate.
    split.
    constructor.
    eapply HTS_alloc ; eauto.
    intros.
    constructor.
    unfold Null. simpl.
    eapply HOS_alloc ; eauto.
    congruence.
    assert (HH := H3 a0).
    revert HH.
    case_eq (h a0).
    intros. destruct o.
    subst_g a a0 ; try congruence.
    destruct (ac a0).
    destruct HH.
    split.
    eapply gHC_ASpecF ; eauto.
    intros.
    case_eq (o f) ; intros.
    simpl.
    subst_g a a2.
    constructor.
    apply False_ind.
    eapply H1 ; eauto.
    generalize (H8 f).
    rewrite H9. simpl.
    eapply gHO_ASpecF ; eauto.
    simpl.
    generalize (H8 f).
    rewrite H9. 
    simpl.     eapply gHO_ASpecF ; eauto.
    auto.
    subst_g a a0 ; try discriminate.
    congruence.
  Qed.
(* end hide *)
    

  Lemma γ_H_ASpecF : forall ac P h',
    γ_H ac (gHC (Ana P)) (gHO (Ana P)) h' -> 
    γ_H ac (gHC (ASpecF P (Ana P))) (gHO (ASpecF P (Ana P))) h'.
   (* begin hide *)
  Proof.
    intros.
    eapply γ_H_mono ; eauto.
    eapply ana_fix ; eauto.
  Qed.
  (* end hide *)

  Lemma γ_E_call : forall P pc0 m cc0 h ac e, 
    nthp P pc0 = Some (Call m) ->
    γ_E ac (gEnv (Ana P) cc0 pc0) e -> 
    γ_H ac (gHC (Ana P)) (gHO (Ana P)) h -> 
    wf_env e h -> 
    γ_E ac
    (gEnv (ASpecF P (Ana P)) (push cc0 pc0 m (abs e ac)) (lk P (dp h e, m))) e.
  (* begin hide *)
  Proof.
    intros.
    unfold γ_E. unfold abs, mem. 
    intros. constructor.
    change (e v ** ac) with ((fun v0 => e v0 ** ac) v).
    eapply ESI_call ; eauto.
    intro.
    unfold wf_env in H2.
    case_eq (e v0) ; simpl.
    intros.
    case_eq (ac a).
    intros. unfold dp. rewrite H3.
    case_eq (h a). destruct o.
    intros.
    generalize (H1 a).
    rewrite H5. rewrite H4. tauto.
    (* contradiction *)
    generalize (H2 v0 a) ; tauto.
    unfold dp. rewrite H3.
    intros.
    generalize (H1 a).
    case_eq (h a).
    destruct o. rewrite H4. tauto.
    rewrite H4. reflexivity.
    unfold dp. intro. rewrite H3. reflexivity.
  Qed.
  (* end hide *)

  Lemma γ_E_ret : forall P pc0 pc' m cc0 e e' ac ac',
    nthp P pc0 = Some Ret -> 
    nthp P pc' = Some (Call m) -> 
    le_ac ac' ac ->
    (∀v : Var, ∀a : Addr, e' v = Some a → ac' a ≠ None) -> 
    γ_E ac' (gEnv (Ana P) cc0 pc') e' -> 
    γ_E ac  (gEnv (Ana P) (push cc0 pc' m (abs e' ac')) pc0) e -> 
    γ_E ac (gEnv (ASpecF P (Ana P)) cc0 (pc' + 1))  (e' [v0 ↦ e v0]).
  (* begin hide *)
  Proof.
    intros.
    unfold γ_E. unfold abs. unfold mem.
    intros. constructor.
    subst_g v0 v.
    change (e v0 ** ac) with (abs e ac v0).
    eapply ESI_ret0  ; eauto.
    change (e' v ** ac) with (abs e' ac v).
    eapply ESI_ret'  ; eauto.
    (**)
    unfold γ_E in *.
    unfold abs in *. clear H4.
    unfold mem in *.
    intros.
    replace (e' v0 ** ac) with (e' v0 ** ac').
    auto.
    case_eq (e' v0) ; intros.
    simpl. unfold le_ac in *.
    case_eq (ac' a) ; intros.
    symmetry. eapply H1 ; eauto.
    apply False_ind. eapply H2 ; eauto.
    reflexivity.
  Qed.
  (* end hide *)

  (* begin hide *)
  Ltac intra_tac :=
    match goal with
      | H : γ ?P (Ana ?P) ?IS |- _ => 
        inv H ; constructor ; 
          [ inv ipattern:Hf ; [ eapply F_init; eauto  | eapply F_cons   ;eauto ;try  eapply γ_F_ASpecF ; eauto ] | idtac]
    end.
  (* end hide *)

  Lemma ISoundness : forall P s, IAcc P s -> γ P (Ana P) s.
  Proof.
    (** The proof is by strong induction over the length of the derivation. *)
    (* begin hide *)
    intros.
    apply IAccIAccN in H.
    destruct H.
    revert H. revert s.
    induction x using Wf_nat.lt_wf_ind.
    intros.
    inv H0.
    (* end hide *)
    (** The base case is solved by [γ_is0] *)
    (* begin hide *) 
    apply  γ_is0. 
    (* end hide *)
    (** The inductive case is by case analysis over the definition of [isos] using the lemma:
       [γ_E_copy], [γ_E_read], [γ_E_write], [γ_H_write], [γ_E_new], [γ_H_new], [γ_E_call], [γ_E_ret]
       and the semantics invariants [IAccNret] , [IAccAc] and [IAcc_wf_H]
       *)
    (* begin hide *)
    assert (Hind :γ P (Ana P) s0).
     eapply (H n) ; eauto.
    apply γ_morph_Ana.
    dup H1 Hacc.
    inv H2;  apply IAccNIAcc in H1.
    inv H4.
    (* copy *)
    intra_tac ; try (eapply γ_E_copy ; eauto); 
      (eapply γ_H_ASpecF ; eauto ; fail).
    (* read *)
    intra_tac ; try (eapply γ_E_read ; eauto); 
      (eapply γ_H_ASpecF ; eauto ; fail).
    (* write *)
    intra_tac;  try (eapply γ_E_write ; eauto).
    eapply γ_F_γ_E in Hf.
    eapply γ_H_write ; eauto.
    (* not new *)
    apply False_ind. apply H3. exists t ; reflexivity.
    (* new *)
    assert (Ha :∀v' : Var, e v' ≠ Some a).
    apply IAcc_wf_H in H1.
    destruct (H1 _ _ _ (refl_equal ([h, ([pc0, e, cc0]) :: s1, ac]))).
    intros. intro. eapply H1 ; eauto. left ; reflexivity.
    assert (HH :le_ac ac (ac [a ↦ Some (alloc cc0 pc0 c (abs e ac))])).
     unfold le_ac.
     intros. subst_g a a0.
     inv Hind.
     generalize (Hh a0).
     rewrite H2. rewrite H3.
     tauto.
    intra_tac.
    eapply γ_E_new ; eauto. 
    eapply le_ac_trans ; eauto.
    eapply γ_E_new ; eauto.
    apply γ_F_γ_E in Hf.
    eapply γ_H_new ; eauto.
    apply IAcc_wf_H in H1.
    destruct (H1 _ _ _ (refl_equal ([h, ([pc0, e, cc0]) :: s1, ac]))).
    auto.
    (* call *)
    inv Hind; constructor.
    eapply F_cons ; eauto. eapply γ_F_ASpecF ; eauto.
    apply le_ac_refl.
    apply γ_F_γ_E in Hf.
    eapply γ_E_call ; eauto.
    apply IAcc_wf_H in H1.
    destruct (H1 _ _ _ (refl_equal ([h, ([pc0, e, cc0]) :: s1, ac]))).  
    eapply H3.
    left ; reflexivity.
    eapply γ_H_ASpecF ; eauto.
    (* return *)
    apply IAccNret in Hacc.
    destruct Hacc as [n' [Hlt [h' [m [Hnth [ac' [Hle [Hpush Hacc]]]]]]]].
    assert (n' < Datatypes.S n) by omega.
    dup Hacc Hacc'.
    apply H in Hacc ; auto.
    apply IAccNIAcc in Hacc'.
    dup Hacc' Hacc2.
    apply IAcc_wf_H in Hacc'.
    assert (forall v a, e' v = (Some a) -> ac' a <> None).
     intros.
     intro.
     rewrite <-  (IAccAc _ _ Hacc2) in H4. simpl in *.
     destruct (Hacc' _ _ _ (refl_equal (MkS h' (MkF pc' e' cc' :: s1) ac'))).
     eapply H6 ; eauto. left;reflexivity.
    inv Hind. inv Hacc. clear H1.
    constructor.
    apply γ_F_γ_E in Hf.
    inv Hf0. constructor.
    eapply γ_E_ret ; eauto.
    eapply F_cons ;eauto.
    eapply γ_F_ASpecF ; eauto.  
    eapply le_ac_trans ; eauto.
    eapply γ_E_ret ; eauto.
    eapply γ_H_ASpecF ; eauto.
  (* end hide *)
  Qed.

  Definition γ_C (P:Program) (A : Atom -> Prop): State -> Prop :=
    fun s => exists is, γ P A is /\ s = EE is.

  Lemma γ_C_mono : forall P A A', A ⊆ A' -> γ_C P A ⊆ γ_C P A'.
  (* begin hide *)
  Proof.
    intros.
    intro. intros.
    unfold γ_C in *.
    destruct H0.
    exists x0.
    intuition.
    eapply γ_mono ; eauto.
  Qed.
  (* end hide *)

  Theorem Soundness : forall P s, Acc P s -> γ_C P (Ana P) s.
  (* begin hide *)
  Proof.
    intros.
    apply AccIAcc in H.
    destruct H. unfold γ_C.
    exists x. split.
    apply ISoundness. tauto.
    tauto.
  Qed.
  (* end hide *)

  (* begin hide *)
  Definition ASpecFIDomKT (P:Program) (cci:CC) (pci:PC) (ei:Var -> option AC) := 
    Build_DomKT _ (ASpecFI P cci pci ei) (ASpecF_mono P cci pci ei) (ASpecF_cont P cci pci ei).

  Definition ASpecFI'DomKT (P:Program) (cci:CC) (pci:PC) (ei:Var -> option AC) := 
    Build_DomKT _ (ASpecFI' P cci pci ei) (ASpecFI'_mono P cci pci ei) (ASpecFI'_cont P cci pci ei).
  (* end hide *)

  (** Lemma [EnvSpecFxTrans] states a transitivity result of [EnvSpecFx]. 
     (This is needed to compare CPA and ooCFA)
     *)
  Lemma EnvSpecFxTrans : forall P HT HO  o pco l l'  pc r a,
    o ∈ EnvSpecFx P  HT HO l pco ->
    EnvSpecIFx P l pco o HT HO l' pc r a ->
    EnvSpecFx P  HT HO l' pc r a.
  Proof.
    (** The proof is by induction over the iterates  of [EnvSpecIFx] *)
    (* begin hide *)
    intros.
    revert H.
    unfold EnvSpecIFx in H0.
    unfold mkCCEnv in H0.
    destruct H0.
    revert H.
    revert l' pc0 r a.
    induction x.
    (* end hide *)
    (** As usual the base case is trivially absurd *)
    (* begin hide *)
    (* bc *)
    simpl. tauto.
    (* end hide*)
    (** For the inductive case, we do a case analysis over [EnvSpecF] *)
    (* begin hide *)
    simpl. intros.
    inv H.
    (* end hide *)
    (** The initial case is trivial verified *)
    (* begin hide *)
    auto.
    (* end hide *)
    (** For the other cases, we use the fact that [EnvSpecFx] is a fixpoint to unroll its definition.
       This allows to apply the induction hypothesis and conclude.
       *)
    (* begin hide *)
    apply EnvSpecFx_pf.
    apply ESI_call with (cc := cc0) (pc:=pc1) ; auto.
    unfold mem. intro.
    apply IHx ; auto. apply H2.
                      (* call / ret *)
    intros.
    apply EnvSpecFx_pf.
    apply ESI_ret0 with (pc:= pc1) (e:=e)(m:= m) ; auto.
    unfold mem. intro.  apply IHx ; auto. apply H3. 
    unfold mem. intro.  apply IHx ; auto. apply H4. 
    apply EnvSpecFx_pf.
    apply ESI_ret' with (m:= m) ; auto.
    unfold mem. intro.  apply IHx ; auto. apply H2. 
                                          (* intra *)
    intros.   apply EnvSpecFx_pf. eapply intra_read_r0 ; eauto. 
    unfold mem. intro.  apply IHx ; auto. apply H2. 
    intros.   apply EnvSpecFx_pf. eapply intra_read_r' ; eauto. 
    unfold mem. intro.  apply IHx ; auto. apply H2. 
    intros.   apply EnvSpecFx_pf. eapply intra_copy_r0 ; eauto. 
    unfold mem. intro.  apply IHx ; auto. apply H2. 
    intros.   apply EnvSpecFx_pf. eapply intra_copy_r' ; eauto. 
    unfold mem. intro.  apply IHx ; auto. apply H3. 
    intros.   apply EnvSpecFx_pf. eapply intra_write  ; eauto. 
    unfold mem. intro.  apply IHx ; auto. apply H2. 
    intros.   apply EnvSpecFx_pf. eapply intra_alloc_r0  ; eauto. 
    unfold mem. intro.  apply IHx ; auto. apply H2. 
    intros.   apply EnvSpecFx_pf. eapply intra_alloc_r'  ; eauto. 
    unfold mem. intro.  apply IHx ; auto. apply H2. 
(* end hide *)
Qed.

End Analysis.


(* Local Variables: *)
(* coding: utf-8 *)
(* coq-prog-name: "/usr/local/bin/coqtop" *)
(* End: *)
